#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
"""Functions for plotting artis estimators and internal structure.

Examples are temperatures, populations, heating/cooling rates.
"""
# import math
import argcomplete
import argparse
import math
import multiprocessing
import os
# import re
import sys
# from collections import namedtuple
# from functools import lru_cache, partial, reduce
# from itertools import chain
from pathlib import Path

import matplotlib.pyplot as plt
# import matplotlib.ticker as ticker
import numpy as np
import pandas as pd
# import scipy.signal
# from astropy import constants as const

import artistools as at
import artistools.initial_composition
import artistools.nltepops
from .estimators import *

colors_tab10 = list(plt.get_cmap('tab10')(np.linspace(0, 1.0, 10)))

# reserve colours for these elements
elementcolors = {
    'Fe': colors_tab10[0],
    'Ni': colors_tab10[1],
    'Co': colors_tab10[2],
}


def get_elemcolor(atomic_number=None, elsymbol=None):
    """Get the colour of an element from the reserved color list (reserving a new one if needed)."""
    assert (atomic_number is None) != (elsymbol is None)
    if atomic_number is not None:
        elsymbol = at.get_elsymbol(atomic_number)

    # assign a new colour to this element if needed

    return elementcolors.setdefault(elsymbol, colors_tab10[len(elementcolors)])


def get_ylabel(variable):
    if variable in variablelongunits:
        return variablelongunits[variable]
    elif variable in variableunits:
        return f'[{variableunits[variable]}]'
    elif variable.split('_')[0] in variableunits:
        return f'[{variableunits[variable.split("_")[0]]}]'
    return ''


def plot_init_abundances(ax, xlist, specieslist, mgilist, modelpath, seriestype, dfalldata=None, args=None,
                         **plotkwargs):
    assert len(xlist) - 1 == len(mgilist)

    if seriestype == 'initabundances':
        mergemodelabundata, _, _ = at.inputmodel.get_modeldata(modelpath, get_abundances=True)
    elif seriestype == 'initmasses':
        mergemodelabundata = at.initial_composition.get_model_abundances_Msun_1D(modelpath)

    for speciesstr in specieslist:
        splitvariablename = speciesstr.split('_')
        elsymbol = splitvariablename[0].strip('0123456789')
        atomic_number = at.get_atomic_number(elsymbol)
        if seriestype == 'initabundances':
            ax.set_ylim(1e-20, 1.0)
            ax.set_ylabel('Initial mass fraction')
            valuetype = 'X_'
        elif seriestype == 'initmasses':
            ax.set_ylabel(r'Initial mass [M$_\odot$]')
            valuetype = 'mass_X_'

        ylist = []
        linelabel = speciesstr
        linestyle = '-'
        for modelgridindex in mgilist:
            if speciesstr.lower() in ['ni_56', 'ni56', '56ni']:
                yvalue = mergemodelabundata.loc[modelgridindex][f'{valuetype}Ni56']
                linelabel = '$^{56}$Ni'
                linestyle = '--'
            elif speciesstr.lower() in ['ni_stb', 'ni_stable']:
                yvalue = (mergemodelabundata.loc[modelgridindex][f'{valuetype}{elsymbol}'] -
                          mergemodelabundata.loc[modelgridindex]['X_Ni56'])
                linelabel = 'Stable Ni'
            elif speciesstr.lower() in ['co_56', 'co56', '56co']:
                yvalue = mergemodelabundata.loc[modelgridindex][f'{valuetype}Co56']
                linelabel = '$^{56}$Co'
            elif speciesstr.lower() in ['fegrp', 'ffegroup']:
                yvalue = mergemodelabundata.loc[modelgridindex][f'{valuetype}Fegroup']
            else:
                yvalue = mergemodelabundata.loc[modelgridindex][f'{valuetype}{elsymbol}']
            ylist.append(yvalue)

        if dfalldata is not None:
            dfalldata['initabundances.' + speciesstr] = ylist

        ylist.insert(0, ylist[0])
        # or ax.step(where='pre', )
        color = get_elemcolor(atomic_number=atomic_number)

        xlist, ylist = apply_filters(xlist, ylist, args)

        ax.plot(xlist, ylist, linewidth=1.5, label=linelabel, linestyle=linestyle, color=color, **plotkwargs)

        # if args.yscale == 'log':
        #     ax.set_yscale('log')


def plot_average_ionisation_excitation(
        ax, xlist, seriestype, params, timestepslist, mgilist,
        estimators, modelpath, dfalldata=None, args=None, **plotkwargs):
    if seriestype == 'averageionisation':
        ax.set_ylabel('Average ion charge')
    elif seriestype == 'averageexcitation':
        ax.set_ylabel('Average excitation [eV]')
    else:
        raise ValueError()

    arr_tdelta = at.get_timestep_times_float(modelpath, loc='delta')
    for paramvalue in params:
        if seriestype == 'averageionisation':
            atomic_number = at.get_atomic_number(paramvalue)
        else:
            atomic_number = at.get_atomic_number(paramvalue.split(' ')[0])
            ion_stage = at.decode_roman_numeral(paramvalue.split(' ')[1])
        ylist = []
        for modelgridindex, timesteps in zip(mgilist, timestepslist):
            valuesum = 0
            tdeltasum = 0
            for timestep in timesteps:

                if seriestype == 'averageionisation':
                    valuesum += (get_averageionisation(
                        estimators[(timestep, modelgridindex)]['populations'], atomic_number) * arr_tdelta[timestep])
                elif seriestype == 'averageexcitation':
                    T_exc = estimators[(timestep, modelgridindex)]['Te']
                    valuesum += (get_averageexcitation(
                        modelpath, modelgridindex, timestep, atomic_number, ion_stage, T_exc) * arr_tdelta[timestep])
                tdeltasum += arr_tdelta[timestep]

            ylist.append(valuesum / tdeltasum)

        color = get_elemcolor(atomic_number=atomic_number)

        if dfalldata is not None:
            dfalldata[seriestype + '.' + paramvalue] = ylist

        ylist.insert(0, ylist[0])

        xlist, ylist = apply_filters(xlist, ylist, args)

        ax.plot(xlist, ylist, label=paramvalue, color=color, **plotkwargs)


def plot_levelpop(
        ax, xlist, seriestype, params, timestepslist, mgilist,
        estimators, modelpath, dfalldata=None, args=None, **plotkwargs):
    if seriestype == 'levelpopulation_dn_on_dvel':
        ax.set_ylabel('dN/dV [{}km$^{{-1}}$ s]')
        ax.yaxis.set_major_formatter(at.ExponentLabelFormatter(ax.get_ylabel(), useMathText=True))
    elif seriestype == 'levelpopulation':
        ax.set_ylabel('X$_{{i}}$ [{}/cm3]')
        ax.yaxis.set_major_formatter(at.ExponentLabelFormatter(ax.get_ylabel(), useMathText=True))
    else:
        raise ValueError()

    modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
    modeldata.eval('modelcellvolume = cellmass_grams / (10 ** logrho)', inplace=True)

    adata = at.atomic.get_levels(modelpath)

    arr_tdelta = at.get_timestep_times_float(modelpath, loc='delta')
    for paramvalue in params:
        paramsplit = paramvalue.split(' ')
        atomic_number = at.get_atomic_number(paramsplit[0])
        ion_stage = at.decode_roman_numeral(paramsplit[1])
        levelindex = int(paramsplit[2])

        ionlevels = adata.query('Z == @atomic_number and ion_stage == @ion_stage').iloc[0].levels
        levelname = ionlevels.iloc[levelindex].levelname
        label = (
            f'{at.get_ionstring(atomic_number, ion_stage, spectral=False)} level {levelindex}:'
            f' {at.nltepops.texifyconfiguration(levelname)}')

        print(f'plot_levelpop {label}')

        # level index query goes outside for caching granularity reasons
        dfnltepops = at.nltepops.read_files(
            modelpath,
            dfquery=f'Z=={atomic_number:.0f} and ion_stage=={ion_stage:.0f}').query('level==@levelindex')

        ylist = []
        for modelgridindex, timesteps in zip(mgilist, timestepslist):
            valuesum = 0
            tdeltasum = 0
            # print(f'modelgridindex {modelgridindex} timesteps {timesteps}')

            for timestep in timesteps:
                levelpop = dfnltepops.query(
                    'modelgridindex==@modelgridindex and timestep==@timestep and Z==@atomic_number'
                    ' and ion_stage==@ion_stage and level==@levelindex').iloc[0].n_NLTE

                valuesum += levelpop * arr_tdelta[timestep]
                tdeltasum += arr_tdelta[timestep]

            if seriestype == 'levelpopulation_dn_on_dvel':
                deltav = modeldata.loc[modelgridindex].velocity_outer - modeldata.loc[modelgridindex].velocity_inner
                ylist.append(valuesum / tdeltasum * modeldata.loc[modelgridindex].modelcellvolume / deltav)
            else:
                ylist.append(valuesum / tdeltasum)

        if dfalldata is not None:
            elsym = at.get_elsymbol(atomic_number).lower()
            if seriestype == 'levelpopulation_dn_on_dvel':
                colname = f'nlevel_on_dv_{elsym}_ionstage{ion_stage}_level{levelindex}'
            else:
                colname = f'nnlevel_{elsym}_ionstage{ion_stage}_level{levelindex}'
            dfalldata[colname] = ylist

        ylist.insert(0, ylist[0])

        xlist, ylist = apply_filters(xlist, ylist, args)

        ax.plot(xlist, ylist, label=label, **plotkwargs)


def plot_multi_ion_series(
        ax, xlist, seriestype, ionlist, timestepslist, mgilist, estimators,
        modelpath, dfalldata=None, args=None, **plotkwargs):
    """Plot an ion-specific property, e.g., populations."""
    assert len(xlist) - 1 == len(mgilist) == len(timestepslist)
    # if seriestype == 'populations':
    #     ax.yaxis.set_major_locator(ticker.MultipleLocator(base=0.10))

    plotted_something = False

    def get_iontuple(ionstr):
        if ionstr in at.elsymbols:
            return (at.get_atomic_number(ionstr), 'ALL')
        elif ' ' in ionstr:
            return (at.get_atomic_number(ionstr.split(' ')[0]), at.decode_roman_numeral(ionstr.split(' ')[1]))
        elif ionstr.rstrip('-0123456789') in at.elsymbols:
            atomic_number = at.get_atomic_number(ionstr.rstrip('-0123456789'))
            return (atomic_number, ionstr)
        else:
            atomic_number = at.get_atomic_number(ionstr.split('_')[0])
            return (atomic_number, ionstr)

    # decoded into atomic number and parameter, e.g., [(26, 1), (26, 2), (26, 'ALL'), (26, 'Fe56')]
    iontuplelist = [get_iontuple(ionstr) for ionstr in ionlist]
    iontuplelist.sort()
    print(f'Subplot with ions: {iontuplelist}')

    missingions = set()
    try:
        if args.classicartis:
            import artistools.estimators.estimators_classic
            compositiondata = artistools.estimators.estimators_classic.get_atomic_composition(modelpath)
        else:
            compositiondata = at.get_composition_data(modelpath)
        for atomic_number, ion_stage in iontuplelist:
            if not hasattr(ion_stage, 'lower') and not args.classicartis and compositiondata.query(
                    'Z == @atomic_number '
                    '& lowermost_ionstage <= @ion_stage '
                    '& uppermost_ionstage >= @ion_stage').empty:
                missingions.add((atomic_number, ion_stage))

    except FileNotFoundError:
        print("WARNING: Could not read an ARTIS compositiondata.txt file")
        for atomic_number, ion_stage in iontuplelist:
            mgits = (timestepslist[0][0], mgilist[0])
            if (atomic_number, ion_stage) not in estimators[mgits]['populations']:
                missingions.add((atomic_number, ion_stage))

    if missingions:
        print(f" Warning: Can't plot {seriestype} for {missingions} "
              f"because these ions are not in compositiondata.txt")

    prev_atomic_number = iontuplelist[0][0]
    colorindex = 0
    for atomic_number, ion_stage in iontuplelist:
        if (atomic_number, ion_stage) in missingions:
            continue

        if atomic_number != prev_atomic_number:
            colorindex += 1

        if seriestype == 'populations':
            if args.ionpoptype == 'absolute':
                ax.set_ylabel('X$_{i}$ [/cm3]')
            elif args.ionpoptype == 'elpop':
                # elsym = at.get_elsymbol(atomic_number)
                ax.set_ylabel(r'X$_{i}$/X$_{\rm element}$')
            elif args.ionpoptype == 'totalpop':
                ax.set_ylabel(r'X$_{i}$/X$_{rm tot}$')
            else:
                assert False
        else:
            ax.set_ylabel(dictlabelreplacements.get(seriestype, seriestype))

        ylist = []
        for modelgridindex, timesteps in zip(mgilist, timestepslist):
            if seriestype == 'populations':
                # if (atomic_number, ion_stage) not in estim['populations']:
                #     print(f'Note: population for {(atomic_number, ion_stage)} not in estimators for '
                #           f'cell {modelgridindex} timesteps {timesteps}')

                try:
                    estimpop = get_averaged_estimators(
                        modelpath, estimators, timesteps, modelgridindex, ['populations'])
                except KeyError:
                    ylist.append(float('nan'))
                    continue

                if ion_stage == 'ALL':
                    nionpop = estimpop.get((atomic_number), 0.)
                elif hasattr(ion_stage, 'lower') and ion_stage.startswith(at.get_elsymbol(atomic_number)):
                    nionpop = estimpop.get(ion_stage, 0.)
                else:
                    nionpop = estimpop.get((atomic_number, ion_stage), 0.)

                try:
                    if args.ionpoptype == 'absolute':
                        yvalue = nionpop  # Plot as fraction of element population
                    elif args.ionpoptype == 'elpop':
                        elpop = estimpop.get(atomic_number, 0.)
                        yvalue = nionpop / elpop  # Plot as fraction of element population
                    elif args.ionpoptype == 'totalpop':
                        totalpop = estimpop['total']
                        yvalue = nionpop / totalpop  # Plot as fraction of total population
                    else:
                        assert False
                except ZeroDivisionError:
                    yvalue = 0.

                ylist.append(yvalue)

            # elif seriestype == 'Alpha_R':
            #     ylist.append(estim['Alpha_R*nne'].get((atomic_number, ion_stage), 0.) / estim['nne'])
            # else:
            #     ylist.append(estim[seriestype].get((atomic_number, ion_stage), 0.))
            else:
                # this is very slow!
                try:
                    estim = get_averaged_estimators(modelpath, estimators, timesteps, modelgridindex, [])
                except KeyError:
                    ylist.append(float('nan'))
                    continue

                dictvars = {}
                for k, value in estim.items():
                    if isinstance(value, dict):
                        dictvars[k] = value.get((atomic_number, ion_stage), 0.)
                    else:
                        dictvars[k] = value

                # dictvars will now define things like 'Te', 'TR',
                # as well as 'populations' which applies to the current ion

                try:
                    yvalue = eval(seriestype, {"__builtins__": math}, dictvars)
                except ZeroDivisionError:
                    yvalue = float('NaN')
                ylist.append(yvalue)

        if hasattr(ion_stage, 'lower') and ion_stage != 'ALL':
            plotlabel = ion_stage
        else:
            plotlabel = at.get_ionstring(atomic_number, ion_stage, spectral=False)

        color = get_elemcolor(atomic_number=atomic_number)

        # linestyle = ['-.', '-', '--', (0, (4, 1, 1, 1)), ':'] + [(0, x) for x in dashes_list][ion_stage - 1]
        if ion_stage == 'ALL':
            dashes = ()
            linewidth = 1.0
        else:
            if hasattr(ion_stage, 'lower') and ion_stage.endswith('stable'):
                index = 8
            elif hasattr(ion_stage, 'lower'):
                # isotopic abundance, use the mass number
                index = int(ion_stage.lstrip(at.get_elsymbol(atomic_number)))
            else:
                index = ion_stage

            dashes_list = [(3, 1, 1, 1), (), (1.5, 1.5), (6, 3), (1, 3)]
            dashes = dashes_list[(index - 1) % len(dashes_list)]
            linewidth_list = [1.0, 1.0, 1.0, 0.7, 0.7]
            linewidth = linewidth_list[(index - 1) % len(linewidth_list)]
            # color = ['blue', 'green', 'red', 'cyan', 'purple', 'grey', 'brown', 'orange'][index - 1]

            if args.colorbyion:
                color = f'C{index - 1 % 10}'
                # plotlabel = f'{at.get_elsymbol(atomic_number)} {at.roman_numerals[ion_stage]}'
                dashes = ()

        # assert colorindex < 10
        # color = f'C{colorindex}'
        # or ax.step(where='pre', )

        if dfalldata is not None:
            elsym = at.get_elsymbol(atomic_number).lower()
            if args.ionpoptype == 'absolute':
                colname = f'nnion_{elsym}_ionstage{ion_stage}'
            elif args.ionpoptype == 'elpop':
                colname = f'nnion_over_nnelem_{elsym}_ionstage{ion_stage}'
            elif args.ionpoptype == 'totalpop':
                colname = f'nnion_over_nntot_{elsym}_ionstage{ion_stage}'
            dfalldata[colname] = ylist

        ylist.insert(0, ylist[0])

        xlist, ylist = apply_filters(xlist, ylist, args)

        ax.plot(xlist, ylist, linewidth=linewidth, label=plotlabel, color=color, dashes=dashes, **plotkwargs)
        prev_atomic_number = atomic_number
        plotted_something = True

    if plotted_something:
        ax.set_yscale(args.yscale)
        if args.yscale == 'log':
            ymin, ymax = ax.get_ylim()
            new_ymax = ymax * 10 ** (0.3 * math.log10(ymax / ymin))
            if ymin > 0 and new_ymax > ymin and np.isfinite(new_ymax):
                ax.set_ylim(ymin, new_ymax)


def plot_series(ax, xlist, variablename, showlegend, timestepslist, mgilist,
                modelpath, estimators, args, nounits=False, dfalldata=None, **plotkwargs):
    """Plot something like Te or TR."""
    assert len(xlist) - 1 == len(mgilist) == len(timestepslist)
    formattedvariablename = dictlabelreplacements.get(variablename, variablename)
    serieslabel = f'{formattedvariablename}'
    if not nounits:
        serieslabel += get_units_string(variablename)

    if showlegend:
        linelabel = serieslabel
    else:
        ax.set_ylabel(serieslabel)
        linelabel = None

    ylist = []
    for modelgridindex, timesteps in zip(mgilist, timestepslist):
        estimavg = get_averaged_estimators(modelpath, estimators, timesteps, modelgridindex, [])
        try:
            ylist.append(eval(variablename, {"__builtins__": math}, estimavg))
        except KeyError:
            if (timesteps[0], modelgridindex) in estimators:
                print(f"Undefined variable: {variablename} in cell {modelgridindex}")
            else:
                print(f'No data for cell {modelgridindex}')
            sys.exit()

    try:
        if math.log10(max(ylist) / min(ylist)) > 2:
            ax.set_yscale('log')
    except ZeroDivisionError:
        ax.set_yscale('log')

    dictcolors = {
        'Te': 'red',
        # 'heating_gamma': 'blue',
        # 'cooling_adiabatic': 'blue'
    }

    # print out the data to stdout. Maybe want to add a CSV export option at some point?
    # print(f'#cellidorvelocity {variablename}\n' + '\n'.join([f'{x}  {y}' for x, y in zip(xlist, ylist)]))

    if dfalldata is not None:
        dfalldata[variablename] = ylist

    ylist.insert(0, ylist[0])

    xlist, ylist = apply_filters(xlist, ylist, args)

    ax.plot(xlist, ylist, linewidth=1.5, label=linelabel, color=dictcolors.get(variablename, None), **plotkwargs)


def get_xlist(xvariable, allnonemptymgilist, estimators, timestepslist, modelpath, args):
    if xvariable in ['cellid', 'modelgridindex']:
        if args.xmax >= 0:
            mgilist_out = [mgi for mgi in allnonemptymgilist if mgi <= args.xmax]
        else:
            mgilist_out = allnonemptymgilist
        xlist = mgilist_out
        timestepslist_out = timestepslist
    elif xvariable == 'timestep':
        mgilist_out = allnonemptymgilist
        xlist = timestepslist
        timestepslist_out = timestepslist
    elif xvariable == 'time':
        mgilist_out = allnonemptymgilist
        timearray = at.get_timestep_times_float(modelpath)
        xlist = [np.mean([timearray[ts] for ts in tslist]) for tslist in timestepslist]
        timestepslist_out = timestepslist
    else:
        xlist = []
        mgilist_out = []
        timestepslist_out = []
        for modelgridindex, timesteps in zip(allnonemptymgilist, timestepslist):
            xvalue = get_averaged_estimators(modelpath, estimators, timesteps, modelgridindex, xvariable)
            xlist.append(xvalue)
            mgilist_out.append(modelgridindex)
            timestepslist_out.append(timesteps)
            if args.xmax > 0 and xvalue > args.xmax:
                break

    xlist, mgilist_out, timestepslist_out = zip(
        *[xmt for xmt in sorted(zip(xlist, mgilist_out, timestepslist_out))])

    assert len(xlist) == len(mgilist_out) == len(timestepslist_out)

    return list(xlist), list(mgilist_out), list(timestepslist_out)


def plot_subplot(ax, timestepslist, xlist, plotitems, mgilist, modelpath,
                 estimators, dfalldata=None, args=None, **plotkwargs):
    """Make plot from ARTIS estimators."""
    # these three lists give the x value, modelgridex, and a list of timesteps (for averaging) for each plot of the plot
    assert len(xlist) - 1 == len(mgilist) == len(timestepslist)
    showlegend = False

    ylabel = 'UNDEFINED'
    sameylabel = True
    for variablename in plotitems:
        if not isinstance(variablename, str):
            pass
        elif ylabel == 'UNDEFINED':
            ylabel = get_ylabel(variablename)
        elif ylabel != get_ylabel(variablename):
            sameylabel = False
            break

    for plotitem in plotitems:
        if isinstance(plotitem, str):
            showlegend = len(plotitems) > 1 or len(variablename) > 20
            plot_series(ax, xlist, plotitem, showlegend, timestepslist, mgilist, modelpath,
                        estimators, args, nounits=sameylabel, dfalldata=dfalldata, **plotkwargs)
            if showlegend and sameylabel:
                ax.set_ylabel(ylabel)
        else:  # it's a sequence of values
            showlegend = True
            seriestype, params = plotitem

            if seriestype in ['initabundances', 'initmasses']:
                plot_init_abundances(ax, xlist, params, mgilist, modelpath, seriestype, dfalldata=dfalldata, args=args)

            elif seriestype == 'levelpopulation' or seriestype.startswith('levelpopulation_'):
                plot_levelpop(
                    ax, xlist, seriestype, params, timestepslist, mgilist, estimators, modelpath,
                    dfalldata=dfalldata, args=args)

            elif seriestype in ['averageionisation', 'averageexcitation']:
                plot_average_ionisation_excitation(
                    ax, xlist, seriestype, params, timestepslist, mgilist, estimators, modelpath,
                    dfalldata=dfalldata, args=args)

            elif seriestype == '_ymin':
                ax.set_ylim(bottom=params)

            elif seriestype == '_ymax':
                ax.set_ylim(top=params)

            elif seriestype == '_yscale':
                ax.set_yscale(params)

            else:
                seriestype, ionlist = plotitem
                plot_multi_ion_series(ax, xlist, seriestype, ionlist, timestepslist, mgilist, estimators,
                                      modelpath, dfalldata, args, **plotkwargs)

    ax.tick_params(right=True)
    if showlegend and not args.nolegend:
        if plotitems[0][0] == 'populations' and args.yscale == 'log':
            ax.legend(loc='best', handlelength=2, ncol=math.ceil(len(plotitems[0][1]) / 2.),
                      frameon=False, numpoints=1)
        else:
            ax.legend(loc='best', handlelength=2,
                      frameon=False, numpoints=1,)  # prop={'size': 9})


def make_plot(modelpath, timestepslist_unfiltered, allnonemptymgilist, estimators, xvariable, plotlist,
              args, **plotkwargs):
    modelname = at.get_model_name(modelpath)
    fig, axes = plt.subplots(nrows=len(plotlist), ncols=1, sharex=True,
                             figsize=(args.figscale * at.figwidth * args.scalefigwidth,
                                      args.figscale * at.figwidth * 0.5 * len(plotlist)),
                             tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})
    if len(plotlist) == 1:
        axes = [axes]

    # ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=5))
    if not args.hidexlabel:
        axes[-1].set_xlabel(f'{xvariable}{get_units_string(xvariable)}')
    xlist, mgilist, timestepslist = get_xlist(
        xvariable, allnonemptymgilist, estimators, timestepslist_unfiltered, modelpath, args)

    dfalldata = pd.DataFrame(index=mgilist)
    dfalldata.index.name = "modelgridindex"
    dfalldata[xvariable] = xlist

    if xvariable.startswith('velocity'):
        xlist = np.insert(xlist, 0, 0.)
    else:
        xlist = np.insert(xlist, 0, xlist[0])

    xmin = args.xmin if args.xmin >= 0 else min(xlist)
    xmax = args.xmax if args.xmax > 0 else max(xlist)

    for ax, plotitems in zip(axes, plotlist):
        ax.set_xlim(left=xmin, right=xmax)
        plot_subplot(ax, timestepslist, xlist, plotitems, mgilist,
                     modelpath, estimators, dfalldata=dfalldata, args=args, **plotkwargs)

    if len(set(mgilist)) == 1 and len(timestepslist[0]) > 1:  # single grid cell versus time plot
        figure_title = f'{modelname}\nCell {mgilist[0]}'

        defaultoutputfile = Path('plotestimators_cell{modelgridindex:03d}.pdf')
        if os.path.isdir(args.outputfile):
            args.outputfile = os.path.join(args.outputfile, defaultoutputfile)

        outfilename = str(args.outputfile).format(modelgridindex=mgilist[0])

    else:
        timeavg = (args.timemin + args.timemax) / 2.
        if args.multiplot and not args.classicartis:
            tdays = estimators[(timestepslist[0][0], mgilist[0])]['tdays']
            figure_title = f'{modelname}\nTimestep {timestepslist[0]} ({tdays:.2f}d)'
        elif args.multiplot and args.classicartis:
            timedays = float(at.get_timestep_time(modelpath, timestepslist[0])[0])
            figure_title = f'{modelname}\nTimestep {timestepslist[0]} ({timedays:.2f}d)'
        else:
            figure_title = f'{modelname}\nTimestep {timestepslist[0]} ({timeavg:.2f}d)'

        defaultoutputfile = Path('plotestimators_ts{timestep:02d}_{timeavg:.0f}d.pdf')
        if os.path.isdir(args.outputfile):
            args.outputfile = os.path.join(args.outputfile, defaultoutputfile)

        outfilename = str(args.outputfile).format(timestep=timestepslist[0][0], timeavg=timeavg)

    if not args.notitle:
        axes[0].set_title(figure_title, fontsize=11)
    # plt.suptitle(figure_title, fontsize=11, verticalalignment='top')

    if args.write_data:
        dfalldata.sort_index(inplace=True)
        dataoutfilename = Path(outfilename).with_suffix('.txt')
        dfalldata.to_csv(dataoutfilename)
        print(f'Saved {dataoutfilename}')

    fig.savefig(outfilename)
    print(f'Saved {outfilename}')

    if args.show:
        plt.show()
    else:
        plt.close()

    return outfilename


def plot_recombrates(modelpath, estimators, atomic_number, ion_stage_list, **plotkwargs):
    fig, axes = plt.subplots(
        nrows=len(ion_stage_list), ncols=1, sharex=True, figsize=(5, 8),
        tight_layout={"pad": 0.5, "w_pad": 0.0, "h_pad": 0.0})
    # ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=5))
    axes[-1].set_xlabel('T_e in kelvins')

    recombcalibrationdata = at.atomic.get_ionrecombratecalibration(modelpath)

    for ax, ion_stage in zip(axes, ion_stage_list):

        ionstr = f'{at.get_elsymbol(atomic_number)} {at.roman_numerals[ion_stage]} to {at.roman_numerals[ion_stage - 1]}'

        listT_e = []
        list_rrc = []
        list_rrc2 = []
        for _, dicttimestepmodelgrid in estimators.items():
            if (not dicttimestepmodelgrid['emptycell']
                    and (atomic_number, ion_stage) in dicttimestepmodelgrid['RRC_LTE_Nahar']):
                listT_e.append(dicttimestepmodelgrid['Te'])
                list_rrc.append(dicttimestepmodelgrid['RRC_LTE_Nahar'][(atomic_number, ion_stage)])
                list_rrc2.append(dicttimestepmodelgrid['Alpha_R'][(atomic_number, ion_stage)])

        if not list_rrc:
            continue

        # sort the pairs by temperature ascending
        listT_e, list_rrc, list_rrc2 = zip(*sorted(zip(listT_e, list_rrc, list_rrc2), key=lambda x: x[0]))

        # markersize=4, marker='s',
        ax.plot(listT_e, list_rrc, linewidth=2, label=f'{ionstr} ARTIS RRC_LTE_Nahar', **plotkwargs)
        ax.plot(listT_e, list_rrc2, linewidth=2, label=f'{ionstr} ARTIS Alpha_R', **plotkwargs)

        try:
            dfrates = recombcalibrationdata[(atomic_number, ion_stage)].query(
                "T_e > @T_e_min & T_e < @T_e_max",
                local_dict={'T_e_min': min(listT_e), 'T_e_max': max(listT_e)})

            ax.plot(dfrates.T_e, dfrates.rrc_total, linewidth=2,
                    label=ionstr + " (calibration)", markersize=6, marker='s', **plotkwargs)
        except KeyError:
            pass

        # rrcfiles = glob.glob(
        #     f'/Users/lshingles/Library/Mobile Documents/com~apple~CloudDocs/GitHub/'
        #     f'artis-atomic/atomic-data-nahar/{at.get_elsymbol(atomic_number).lower()}{ion_stage - 1}.rrc*.txt')
        # if rrcfiles:
        #     dfrecombrates = get_ionrecombrates_fromfile(rrcfiles[0])
        #
        #     dfrecombrates.query("logT > @logT_e_min & logT < @logT_e_max",
        #                         local_dict={'logT_e_min': math.log10(min(listT_e)),
        #                                     'logT_e_max': math.log10(max(listT_e))}, inplace=True)
        #
        #     listT_e_Nahar = [10 ** x for x in dfrecombrates['logT'].values]
        #     ax.plot(listT_e_Nahar, dfrecombrates['RRC_total'], linewidth=2,
        #             label=ionstr + " (Nahar)", markersize=6, marker='s', **plotkwargs)

        ax.legend(loc='best', handlelength=2, frameon=False, numpoints=1, prop={'size': 10})

    # modelname = at.get_model_name(".")
    # plotlabel = f'Timestep {timestep}'
    # time_days = float(at.get_timestep_time('spec.out', timestep))
    # if time_days >= 0:
    #     plotlabel += f' (t={time_days:.2f}d)'
    # fig.suptitle(plotlabel, fontsize=12)
    elsymbol = at.get_elsymbol(atomic_number)
    outfilename = f"plotestimators_recombrates_{elsymbol}.pdf"
    fig.savefig(outfilename)
    print(f'Saved {outfilename}')
    plt.close()


def addargs(parser):
    parser.add_argument('-modelpath', default='.',
                        help='Paths to ARTIS folder (or virtual path e.g. codecomparison/ddc10/cmfgen)')

    parser.add_argument('--recombrates', action='store_true',
                        help='Make a recombination rate plot')

    parser.add_argument('-modelgridindex', '-cell', '-mgi', type=int, default=-1,
                        help='Modelgridindex for time evolution plot')

    parser.add_argument('-timestep', '-ts', nargs='?',
                        help='Timestep number for internal structure plot')

    parser.add_argument('-timedays', '-time', '-t', nargs='?',
                        help='Time in days to plot for internal structure plot')

    parser.add_argument('-timemin', type=float,
                        help='Lower time in days')

    parser.add_argument('-timemax', type=float,
                        help='Upper time in days')

    parser.add_argument('--multiplot', action='store_true',
                        help='Make multiple plots for timesteps in range')

    parser.add_argument('-x',
                        help='Horizontal axis variable, e.g. cellid, velocity, timestep, or time')

    parser.add_argument('-xmin', type=float, default=-1,
                        help='Plot range: minimum x value')

    parser.add_argument('-xmax', type=float, default=-1,
                        help='Plot range: maximum x value')

    parser.add_argument('-yscale', default='log', choices=['log', 'linear'],
                        help='Set yscale to log or linear (default log)')

    parser.add_argument('--hidexlabel', action='store_true',
                        help='Hide the bottom horizontal axis label')

    parser.add_argument('-filtermovingavg', type=int, default=0,
                        help='Smoothing length (1 is same as none)')

    parser.add_argument('-filtersavgol', nargs=2,
                        help='Savitzky–Golay filter. Specify the window_length and polyorder.'
                        'e.g. -filtersavgol 5 3')

    parser.add_argument('--notitle', action='store_true',
                        help='Suppress the top title from the plot')

    parser.add_argument('-plotlist', type=list, default=[],
                        help='Plot list (when calling from Python only)')

    parser.add_argument('-ionpoptype', default='elpop', choices=['absolute', 'totalpop', 'elpop'],
                        help=(
                            'Plot absolute ion populations, or ion populations as a'
                            ' fraction of total or element population'))

    parser.add_argument('--nolegend', action='store_true',
                        help='Suppress the legend from the plot')

    parser.add_argument('-figscale', type=float, default=1.,
                        help='Scale factor for plot area. 1.0 is for single-column')

    parser.add_argument('-scalefigwidth', type=float, default=1.,
                        help='Scale factor for plot width.')

    parser.add_argument('--show', action='store_true',
                        help='Show plot before quitting')

    parser.add_argument('--write_data', action='store_true',
                        help='Save data used to generate the plot in a CSV file')

    parser.add_argument('-o', action='store', dest='outputfile', type=Path, default=Path(),
                        help='Filename for PDF file')

    parser.add_argument('--colorbyion', action='store_true',
                        help='Populations plots colored by ion rather than element')

    parser.add_argument('--classicartis', action='store_true',
                        help='Flag to show using output from classic ARTIS branch')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot ARTIS estimators.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    modelpath = Path(args.modelpath)

    modelname = at.get_model_name(modelpath)

    if not args.timedays and not args.timestep and args.modelgridindex > -1:
        args.timestep = f'0-{len(at.get_timestep_times_float(modelpath)) - 1}'

    (timestepmin, timestepmax, args.timemin, args.timemax) = at.get_time_range(
         modelpath, args.timestep, args.timemin, args.timemax, args.timedays)

    print(f"Plotting estimators for '{modelname}' timesteps {timestepmin} to {timestepmax} "
          f"({args.timemin:.1f} to {args.timemax:.1f}d)")

    timesteps_included = list(range(timestepmin, timestepmax + 1))

    if args.classicartis:
        import artistools.estimators.estimators_classic
        modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
        estimators = artistools.estimators.estimators_classic.read_classic_estimators(modelpath, modeldata)
    else:
        estimators = read_estimators(modelpath=modelpath, modelgridindex=args.modelgridindex, timestep=tuple(timesteps_included))

    for ts in reversed(timesteps_included):
        tswithdata = [ts for (ts, mgi) in estimators.keys()]
        for ts in timesteps_included:
            if ts not in tswithdata:
                timesteps_included.remove(ts)
                print(f"ts {ts} requested but no data found. Removing.")

    if not timesteps_included:
        print("No timesteps with data are included")
        return

    if args.plotlist:
        plotlist = args.plotlist
    else:
        plotlist = [
            [['initabundances', ['Fe', 'Ni_stable', 'Ni_56']]],
            ['heating_dep', 'heating_coll', 'heating_bf', 'heating_ff',
             ['_yscale', 'linear']],
            ['cooling_adiabatic', 'cooling_coll', 'cooling_fb', 'cooling_ff',
             ['_yscale', 'linear']],
            # [['initmasses', ['Ni_56', 'He', 'C', 'Mg']]],
            # ['heating_gamma/gamma_dep'],
            # ['nne'],
            ['W', 'TR', 'Te', 'TJ',
             ['_yscale', 'linear']],
            # ['Te'],
            # [['averageionisation', ['Fe', 'Ni']]],
            # [['averageexcitation', ['Fe II', 'Fe III']]],
            # [['populations', ['Sr89', 'Sr90', 'Sr91', 'Sr92', 'Sr93', 'Sr94', 'Sr95']],
            #  ['_ymin', 1e-3], ['_ymax', 5]],
            # [['populations', ['Fe', 'Co', 'Ni', 'Sr']]],
            # [['populations', ['He I', 'He II', 'He III']]],
            # [['populations', ['C I', 'C II', 'C III', 'C IV', 'C V']]],
            # [['populations', ['O I', 'O II', 'O III', 'O IV']]],
            # [['populations', ['Ne I', 'Ne II', 'Ne III', 'Ne IV', 'Ne V']]],
            # [['populations', ['Si I', 'Si II', 'Si III', 'Si IV', 'Si V']]],
            # [['populations', ['Cr I', 'Cr II', 'Cr III', 'Cr IV', 'Cr V']]],
            [['populations', ['Fe I', 'Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Fe VI', 'Fe VII', 'Fe VIII']]],
            [['populations', ['Co I', 'Co II', 'Co III', 'Co IV', 'Co V', 'Co VI', 'Co VII']]],
            [['populations', ['Ni I', 'Ni II', 'Ni III', 'Ni IV', 'Ni V', 'Ni VI', 'Ni VII']]],
            # [['populations', ['Fe II', 'Fe III', 'Co II', 'Co III', 'Ni II', 'Ni III']]],
            # [['populations', ['Fe I', 'Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Ni II']]],
            # [['RRC_LTE_Nahar', ['Fe II', 'Fe III', 'Fe IV', 'Fe V']]],
            # [['RRC_LTE_Nahar', ['Co II', 'Co III', 'Co IV', 'Co V']]],
            # [['RRC_LTE_Nahar', ['Ni I', 'Ni II', 'Ni III', 'Ni IV', 'Ni V', 'Ni VI', 'Ni VII']]],
            # [['Alpha_R / RRC_LTE_Nahar', ['Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Ni III']]],
            # [['gamma_NT', ['Fe I', 'Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Ni II']]],
        ]

    if args.recombrates:
        plot_recombrates(modelpath, estimators, 26, [2, 3, 4, 5])
        plot_recombrates(modelpath, estimators, 27, [3, 4])
        plot_recombrates(modelpath, estimators, 28, [3, 4, 5])
    else:
        modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)

        if args.modelgridindex > -1 or args.x in ['time', 'timestep']:
            # plot time evolution in specific cell
            if not args.x:
                args.x = 'time'
            mgilist = [args.modelgridindex] * len(timesteps_included)
            timesteplist_unfiltered = [(ts,) for ts in timesteps_included]
            make_plot(modelpath, timesteplist_unfiltered, mgilist, estimators, args.x, plotlist, args)
        else:
            # plot a range of cells in a time snapshot showing internal structure

            if not args.x:
                args.x = 'velocity_outer'

            if args.classicartis:
                allnonemptymgilist = [modelgridindex for modelgridindex in modeldata.index
                                      if (timesteps_included[0], modelgridindex) in estimators]
            else:
                allnonemptymgilist = [modelgridindex for modelgridindex in modeldata.index
                                      if not estimators[(timesteps_included[0], modelgridindex)]['emptycell']]

            if args.multiplot:
                pdf_list = []
                modelpath_list = []
                for timestep in range(timestepmin, timestepmax + 1):
                    timesteplist_unfiltered = [[timestep]] * len(allnonemptymgilist)
                    outfilename = make_plot(modelpath, timesteplist_unfiltered, allnonemptymgilist, estimators, args.x,
                                            plotlist, args)

                    if '/' in outfilename:
                        outfilename = outfilename.split('/')[1]

                    pdf_list.append(outfilename)
                    modelpath_list.append(modelpath)

                if len(pdf_list) > 1:
                    at.join_pdf_files(pdf_list, modelpath_list)

            else:
                timesteplist_unfiltered = [timesteps_included] * len(allnonemptymgilist)
                make_plot(modelpath, timesteplist_unfiltered, allnonemptymgilist, estimators, args.x, plotlist, args)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



#!/usr/bin/env python3

import argparse
import math
import os.path

import numpy as np
import pandas as pd
from astropy import units as u

import artistools as at
import artistools.inputmodel


def min_dist(listin, number):
    """Return the minimum distance between number and any item in listin."""
    min_dist_found = -1

    for x in listin:
        dist = abs(x - number)
        if dist < min_dist_found or min_dist_found < 0:
            min_dist_found = dist

    return min_dist_found


def addargs(parser):
    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path for output files')


def main(args=None, argsraw=None, **kwargs) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Create Botyanski et al. 2017 .')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    e_k = 1.2  # in units of 10^51 erg
    m_ej = 1.4  # in solar masses
    x_stb = 0.05  # mass fraction of stable Fe54 and Ni58 in Ni56 zone
    t_model_init_days = 0.0002
    t200 = t_model_init_days / 200  # time in units of 200 days

    delta = 0
    n = 10

    # density transition
    v_transition = 10943 * e_k ** 0.5 * m_ej ** -0.5  # km/s
    rho_0 = 4.9e-17 * (e_k ** -1.5) * (m_ej ** 2.5) * (t200 ** -3)  # g cm^-3

    print(f'v_transition = {v_transition:.3f}')

    # composition transition from Ni56-rich to IME-rich
    mni56 = 0.6 * u.solMass
    volni56 = (mni56 / ((1 - x_stb) * rho_0 * u.g * u.cm ** -3)).to('cm3')
    rni56 = (3 / 4 / math.pi * volni56) ** (1 / 3.)
    v_ni56 = (rni56 / (200 * t200 * u.day)).to('km/s').value

    r = (v_ni56 * (u.km / u.s) * 200 * t200 * u.day).to('cm')
    m = (4 * math.pi / 3 * (r ** 3) * (rho_0 * u.g * u.cm ** -3)).to('solMass')
    print(f'Ni56 region outer velocity = {v_ni56:.3f}, M={m:.3f}')

    dfmodel = pd.DataFrame(
        columns=['inputcellid', 'velocity_outer', 'logrho', 'X_Fegroup', 'X_Ni56', 'X_Co56', 'X_Fe52', 'X_Cr48'])
    dfmodel.index.name = 'cellid'
    dfelabundances = pd.DataFrame(columns=['inputcellid', *['X_' + at.get_elsymbol(x) for x in range(1, 31)]])
    dfelabundances.index.name = 'cellid'

    fixed_points = [v_transition, v_ni56]
    regular_points = [v for v in np.arange(0, 14500, 1000)[1:] if min_dist(fixed_points, v) > 200]
    vlist = sorted(list([*fixed_points, *regular_points]))

    v_inner = 0.  # velocity at inner boundary of cell
    m_tot = 0.
    for cellid, v_outer in enumerate(vlist):  # km / s
        rho = rho_0 * (0.5 * (v_inner + v_outer) / v_transition) ** -(delta if v_outer <= v_transition else n)
        abundances = [0. for _ in range(31)]
        if v_outer <= v_ni56:
            # Ni56-rich zone
            radioabundances = [1.0, 0.95, 0.0, 0.0, 0.0]
            abundances[26] = 0.025
            abundances[28] = 0.975
        else:
            # Intermediate-mass elements
            radioabundances = [0.0, 0.0, 0.0, 0.0, 0.0]
            abundances[14] = 0.7
            abundances[16] = 0.29
            abundances[20] = 0.01

        dfmodel.loc[cellid] = [cellid + 1, v_outer, math.log10(rho), *radioabundances]
        dfelabundances.loc[cellid] = [cellid + 1, *abundances[1:31]]
        r_inner, r_outer = [(v * u.km / u.s * t200 * 200 * u.day).to('cm').value for v in [v_inner, v_outer]]

        vol_shell = 4 * math.pi / 3 * (r_outer ** 3 - r_inner ** 3)
        m_shell = rho * vol_shell / u.solMass.to('g')
        m_tot += m_shell

        v_inner = v_outer
    print(f'M_tot = {m_tot:.3f} solMass')

    at.inputmodel.save_modeldata(dfmodel, t_model_init_days, os.path.join(args.outputpath, 'model.txt'))
    at.inputmodel.save_initialabundances(dfelabundances, os.path.join(args.outputpath, 'abundances.txt'))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
# import glob
import math
import multiprocessing
# import re
from collections import namedtuple
from pathlib import Path

import matplotlib.pyplot as plt
# import numexpr as ne
import numpy as np
import pandas as pd
from astropy import constants as const
# from astropy import units as u

import artistools as at
import artistools.estimators
import artistools.nltepops
import artistools.spectra

defaultoutputfile = 'plottransitions_cell{cell:03d}_ts{timestep:02d}_{time_days:.0f}d.pdf'

iontuple = namedtuple('iontuple', 'Z ion_stage')


def get_kurucz_transitions():
    hc_evcm = (const.h * const.c).to('eV cm').value
    transitiontuple = namedtuple('transition',
                                 'Z ionstage lambda_angstroms A lower_energy_ev upper_energy_ev lower_g upper_g')
    translist = []
    ionlist = []
    with open('gfall.dat', 'r') as fnist:
        for line in fnist:
            row = line.split()
            if len(row) >= 24:
                Z, ionstage = int(row[2].split('.')[0]), int(row[2].split('.')[1]) + 1
                if Z < 44 or ionstage >= 2:  # and Z not in [26, 27]
                    continue
                lambda_angstroms = float(line[:12]) * 10
                loggf = float(line[11:18])
                lower_energy_ev, upper_energy_ev = hc_evcm * float(line[24:36]), hc_evcm * float(line[52:64])
                lower_g, upper_g = 2 * float(line[36:42]) + 1, 2 * float(line[64:70]) + 1
                fij = (10 ** loggf) / lower_g
                A = fij / (1.49919e-16 * upper_g / lower_g * lambda_angstroms ** 2)
                translist.append(transitiontuple(
                    Z, ionstage, lambda_angstroms, A, lower_energy_ev, upper_energy_ev, lower_g, upper_g))

                if iontuple(Z, ionstage) not in ionlist:
                    ionlist.append(iontuple(Z, ionstage))

    dftransitions = pd.DataFrame(translist, columns=transitiontuple._fields)
    return dftransitions, ionlist


def get_nist_transitions(filename):
    transitiontuple = namedtuple('transition', 'lambda_angstroms A lower_energy_ev upper_energy_ev lower_g upper_g')
    translist = []
    with open(filename, 'r') as fnist:
        for line in fnist:
            row = line.split('|')
            if len(row) == 17 and '-' in row[5]:
                if len(row[0].strip()) > 0:
                    lambda_angstroms = float(row[0])
                elif len(row[1].strip()) > 0:
                    lambda_angstroms = float(row[1])
                else:
                    continue
                if len(row[3].strip()) > 0:
                    A = float(row[3])
                else:
                    # continue
                    A = 1e8
                lower_energy_ev, upper_energy_ev = [float(x.strip(' []')) for x in row[5].split('-')]
                lower_g, upper_g = [float(x.strip()) for x in row[12].split('-')]
                translist.append(transitiontuple(
                    lambda_angstroms, A, lower_energy_ev, upper_energy_ev, lower_g, upper_g))

    dftransitions = pd.DataFrame(translist, columns=transitiontuple._fields)
    return dftransitions


def generate_ion_spectrum(transitions, xvalues, popcolumn, plot_resolution, args):
    yvalues = np.zeros(len(xvalues))

    # iterate over lines
    for _, line in transitions.iterrows():
        flux = line['flux_factor'] * line[popcolumn]

        # contribute the Gaussian line profile to the discrete flux bins

        centre_index = int(round((line['lambda_angstroms'] - args.xmin) / plot_resolution))
        sigma_angstroms = line['lambda_angstroms'] * args.sigma_v / const.c.to('km / s').value
        sigma_gridpoints = int(math.ceil(sigma_angstroms / plot_resolution))
        window_left_index = max(int(centre_index - args.gaussian_window * sigma_gridpoints), 0)
        window_right_index = min(int(centre_index + args.gaussian_window * sigma_gridpoints), len(xvalues))

        for x in range(max(0, window_left_index), min(len(xvalues), window_right_index)):
            yvalues[x] += flux * math.exp(
                -((x - centre_index) * plot_resolution / sigma_angstroms) ** 2) / sigma_angstroms

    return yvalues


def make_plot(
        xvalues, yvalues, temperature_list, vardict, ionlist, ionpopdict, xmin, xmax, figure_title, outputfilename):
    # npanels = len(ionlist) + 1
    npanels = len(ionlist)

    fig, axes = plt.subplots(
        nrows=npanels, ncols=1, sharex=True, sharey=False, figsize=(6, 2 * (len(ionlist) + 1)),
        tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    if len(ionlist) == 1:
        axes = [axes]

    if figure_title:
        print(figure_title)
        axes[0].set_title(figure_title, fontsize=10)

    peak_y_value = -1
    yvalues_combined = np.zeros((len(temperature_list), len(xvalues)))
    for seriesindex, temperature in enumerate(temperature_list):
        T_exc = eval(temperature, vardict)
        serieslabel = 'NLTE' if T_exc < 0 else f'LTE {temperature} = {T_exc:.0f} K'
        for ion_index, axis in enumerate(axes[:len(ionlist)]):
            # an ion subplot
            yvalues_combined[seriesindex] += yvalues[seriesindex][ion_index]

            axis.plot(xvalues, yvalues[seriesindex][ion_index], linewidth=1.5, label=serieslabel)

            peak_y_value = max(yvalues[seriesindex][ion_index])

            axis.legend(loc='upper left', handlelength=1, frameon=False, numpoints=1, prop={'size': 8})

        if len(axes) > len(ionlist):
            axes[len(ionlist)].plot(xvalues, yvalues_combined[seriesindex], linewidth=1.5, label=serieslabel)
            peak_y_value = max(peak_y_value, max(yvalues_combined[seriesindex]))

    axislabels = [
        f'{at.get_elsymbol(Z)} {at.roman_numerals[ion_stage]}\n(pop={ionpopdict[(Z, ion_stage)]:.1e}/cm3)'
        for (Z, ion_stage) in ionlist]
    axislabels += ['Total']

    for axis, axislabel in zip(axes, axislabels):
        axis.annotate(
            axislabel, xy=(0.99, 0.96), xycoords='axes fraction',
            horizontalalignment='right', verticalalignment='top', fontsize=10)

    # at.spectra.plot_reference_spectrum(
    #     'dop_dered_SN2013aa_20140208_fc_final.txt', axes[-1], xmin, xmax, True,
    #     scale_to_peak=peak_y_value, zorder=-1, linewidth=1, color='black')
    #
    # at.spectra.plot_reference_spectrum(
    #     '2003du_20031213_3219_8822_00.txt', axes[-1], xmin, xmax,
    #     scale_to_peak=peak_y_value, zorder=-1, linewidth=1, color='black')

    axes[-1].set_xlabel(r'Wavelength ($\AA$)')

    for axis in axes:
        axis.set_xlim(xmin, xmax)
        axis.set_ylabel(r'$\propto$ F$_\lambda$')

    print(f"Saving '{outputfilename}'")
    fig.savefig(outputfilename, format='pdf')
    plt.close()


def add_upper_lte_pop(dftransitions, T_exc, ionpop, ltepartfunc, columnname=None):
    K_B = const.k_B.to('eV / K').value
    scalefactor = ionpop / ltepartfunc
    if columnname is None:
        columnname = f'upper_pop_lte_{T_exc:.0f}K'
    dftransitions.eval(
        f'{columnname} = @scalefactor * upper_g * exp(-upper_energy_ev / @K_B / @T_exc)',
        inplace=True)


def addargs(parser):
    parser.add_argument('-modelpath', default=None, type=Path,
                        help='Path to ARTIS folder')

    parser.add_argument('-xmin', type=int, default=3500,
                        help='Plot range: minimum wavelength in Angstroms')

    parser.add_argument('-xmax', type=int, default=8000,
                        help='Plot range: maximum wavelength in Angstroms')

    parser.add_argument('-T', type=float, dest='T', default=[], nargs='*',
                        help='Temperature in Kelvin')

    parser.add_argument('-sigma_v', type=float, default=5500.,
                        help='Gaussian width in km/s')

    parser.add_argument('-gaussian_window', type=float, default=3,
                        help='Truncate Gaussian line profiles n sigmas from the centre')

    parser.add_argument('--include-permitted', action='store_true',
                        help='Also consider permitted lines')

    parser.add_argument('-timedays', '-time', '-t',
                        help='Time in days to plot')

    parser.add_argument('-timestep', '-ts', type=int, default=70,
                        help='Timestep number to plot')

    parser.add_argument('-modelgridindex', '-cell', type=int, default=0,
                        help='Modelgridindex to plot')

    parser.add_argument('--normalised', action='store_true',
                        help='Normalise all spectra to their peak values')

    parser.add_argument('--print-lines', action='store_true',
                        help='Output details of matching line details to standard out')

    parser.add_argument('--atomicdatabase', default='artis', choices=['artis', 'kurucz', 'nist'],
                        help='Source of atomic data for excitation transitions')

    parser.add_argument('-o', action='store', dest='outputfile',
                        default=defaultoutputfile,
                        help='path/filename for PDF file')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot estimated spectra from bound-bound transitions.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if Path(args.outputfile).is_dir():
        args.outputfile = Path(args.outputfile, defaultoutputfile)

    if args.modelpath:
        from_model = True
    else:
        from_model = False
        args.modelpath = Path()

    modelpath = args.modelpath
    if from_model:
        modelgridindex = args.modelgridindex

        if args.timedays:
            timestep = at.get_timestep_of_timedays(modelpath, args.timedays)
        else:
            timestep = args.timestep

        modeldata, _, _ = at.inputmodel.get_modeldata(Path(modelpath, 'model.txt'))
        estimators_all = at.estimators.read_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)
        if not estimators_all:
            return -1

        estimators = estimators_all[(timestep, modelgridindex)]
        if estimators['emptycell']:
            print(f'ERROR: cell {modelgridindex} is marked as empty')
            return -1

    # also calculate wavelengths outside the plot range to include lines whose
    # edges pass through the plot range
    plot_xmin_wide = args.xmin * (1 - args.gaussian_window * args.sigma_v / const.c.to('km / s').value)
    plot_xmax_wide = args.xmax * (1 + args.gaussian_window * args.sigma_v / const.c.to('km / s').value)

    ionlist = [
        (26, 1), (26, 2), (26, 3),
        (27, 2), (27, 3),
        (28, 2), (28, 3),
        # (28, 2),
        # iontuple(45, 1),
        # iontuple(54, 1),
        # iontuple(54, 2),
        # iontuple(55, 1),
        # iontuple(55, 2),
        # iontuple(58, 1),
        # iontuple(79, 1),
        # iontuple(83, 1),
        # iontuple(26, 2),
        # iontuple(26, 3),
    ]

    if args.atomicdatabase == 'kurucz':
        dftransgfall, ionlist = get_kurucz_transitions()

    ionlist.sort()

    # resolution of the plot in Angstroms
    plot_resolution = max(1, int((args.xmax - args.xmin) / 1000))

    if args.atomicdatabase == 'artis':
        adata = at.atomic.get_levels(modelpath, tuple(ionlist), get_transitions=True)

    if from_model:
        dfnltepops = at.nltepops.read_files(modelpath, modelgridindex=modelgridindex, timestep=timestep)

        if dfnltepops is None or dfnltepops.empty:
            print(f'ERROR: no NLTE populations for cell {modelgridindex} at timestep {timestep}')
            return -1

        ionpopdict = {(Z, ion_stage): dfnltepops.query(
            'Z==@Z and ion_stage==@ion_stage')['n_NLTE'].sum() for Z, ion_stage in ionlist}

        modelname = at.get_model_name(modelpath)
        velocity = modeldata['velocity_outer'][modelgridindex]

        Te = estimators['Te']
        TR = estimators['TR']
        figure_title = f'{modelname}\n'
        figure_title += (f'Cell {modelgridindex} ({velocity} km/s) with '
                         f'Te = {Te:.1f} K, TR = {TR:.1f} K at timestep {timestep}')
        time_days = float(at.get_timestep_time(modelpath, timestep))
        if time_days != -1:
            figure_title += f' ({time_days:.1f}d)'

        # -1 means use NLTE populations
        temperature_list = ['Te', 'TR', '-1']
        temperature_list = ['-1']
        vardict = {'Te': Te, 'TR': TR}
    else:
        if not args.T:
            args.T = [2000]
        if len(args.T) == 1:
            figure_title = f'Te = {args.T[0]:.1f}'
        else:
            figure_title = None

        temperature_list = []
        vardict = {}
        for index, temperature in enumerate(args.T):
            tlabel = 'Te'
            if index > 0:
                tlabel += f'_{index + 1}'
            vardict[tlabel] = temperature
            temperature_list.append(tlabel)

        # Fe3overFe2 = 8  # number ratio
        # ionpopdict = {
        #     (26, 2): 1 / (1 + Fe3overFe2),
        #     (26, 3): Fe3overFe2 / (1 + Fe3overFe2),
        #     (28, 2): 1.0e-2,
        # }
        ionpopdict = {ion: 1 for ion in ionlist}

    hc = (const.h * const.c).to('eV Angstrom').value

    xvalues = np.arange(args.xmin, args.xmax, step=plot_resolution)
    yvalues = np.zeros((len(temperature_list) + 1, len(ionlist), len(xvalues)))

    for _, ion in adata.iterrows() if args.atomicdatabase == 'artis' else enumerate(ionlist):
        ionid = (ion.Z, ion.ion_stage)
        if ionid not in ionlist:
            continue
        else:
            ionindex = ionlist.index(ionid)

        if args.atomicdatabase == 'kurucz':
            dftransitions = dftransgfall.query('Z == @ion.Z and ionstage == @ion.ion_stage', inplace=False).copy()
        elif args.atomicdatabase == 'nist':
            dftransitions = get_nist_transitions(f'nist/nist-{ion.Z:02d}-{ion.ion_stage:02d}.txt')
        else:
            dftransitions = ion.transitions

        print(f'\n======> {at.get_elsymbol(ion.Z)} {at.roman_numerals[ion.ion_stage]:3s} '
              f'(pop={ionpopdict[ionid]:.2e} / cm3, {len(dftransitions):6d} transitions)')

        if not args.include_permitted and not dftransitions.empty:
            dftransitions.query('forbidden == True', inplace=True)
            print(f'  ({len(ion.transitions):6d} forbidden)')

        if not dftransitions.empty:
            if args.atomicdatabase == 'artis':
                dftransitions.eval('upper_energy_ev = @ion.levels.loc[upper].energy_ev.values', inplace=True)
                dftransitions.eval('lower_energy_ev = @ion.levels.loc[lower].energy_ev.values', inplace=True)
                dftransitions.eval('lambda_angstroms = @hc / (upper_energy_ev - lower_energy_ev)', inplace=True)

            dftransitions.query('lambda_angstroms >= @plot_xmin_wide & lambda_angstroms <= @plot_xmax_wide',
                                inplace=True)

            dftransitions.sort_values(by='lambda_angstroms', inplace=True)

            print(f'  {len(dftransitions)} plottable transitions')

            if args.atomicdatabase == 'artis':
                dftransitions.eval('upper_g = @ion.levels.loc[upper].g.values', inplace=True)
                K_B = const.k_B.to('eV / K').value
                T_exc = vardict['Te']
                ltepartfunc = ion.levels.eval('g * exp(-energy_ev / @K_B / @T_exc)').sum()
            else:
                ltepartfunc = 1.0

            dftransitions.eval('flux_factor = (upper_energy_ev - lower_energy_ev) * A', inplace=True)
            add_upper_lte_pop(dftransitions, vardict['Te'], ionpopdict[ionid], ltepartfunc, columnname='upper_pop_Te')

            for seriesindex, temperature in enumerate(temperature_list):
                T_exc = eval(temperature, vardict)
                if T_exc < 0:
                    dfnltepops_thision = dfnltepops.query('Z==@ion.Z & ion_stage==@ion.ion_stage')

                    nltepopdict = {x.level: x['n_NLTE'] for _, x in dfnltepops_thision.iterrows()}

                    dftransitions['upper_pop_nlte'] = dftransitions.apply(
                        lambda x: nltepopdict.get(x.upper, 0.), axis=1)

                    # dftransitions['lower_pop_nlte'] = dftransitions.apply(
                    #     lambda x: nltepopdict.get(x.lower, 0.), axis=1)

                    popcolumnname = 'upper_pop_nlte'
                    dftransitions.eval(f'flux_factor_nlte = flux_factor * {popcolumnname}', inplace=True)
                    dftransitions.eval('upper_departure = upper_pop_nlte / upper_pop_Te', inplace=True)
                    if ionid == (26, 2):
                        fe2depcoeff = dftransitions.query('upper == 16 and lower == 5').iloc[0].upper_departure
                    elif ionid == (28, 2):
                        ni2depcoeff = dftransitions.query('upper == 6 and lower == 0').iloc[0].upper_departure

                    with pd.option_context('display.width', 200):
                        print(dftransitions.nlargest(1, 'flux_factor_nlte'))
                else:
                    popcolumnname = f'upper_pop_lte_{T_exc:.0f}K'
                    if args.atomicdatabase == 'artis':
                        dftransitions.eval('upper_g = @ion.levels.loc[upper].g.values', inplace=True)
                        K_B = const.k_B.to('eV / K').value
                        ltepartfunc = ion.levels.eval('g * exp(-energy_ev / @K_B / @T_exc)').sum()
                    else:
                        ltepartfunc = 1.0
                    add_upper_lte_pop(dftransitions, T_exc, ionpopdict[ionid], ltepartfunc, columnname=popcolumnname)

                if args.print_lines:
                    dftransitions.eval(f'flux_factor_{popcolumnname} = flux_factor * {popcolumnname}', inplace=True)

                yvalues[seriesindex][ionindex] = generate_ion_spectrum(dftransitions, xvalues,
                                                                       popcolumnname, plot_resolution, args)
                if args.normalised:
                    yvalues[seriesindex][ionindex] /= max(yvalues[seriesindex][ionindex])  # todo: move to ax.plot line

        if args.print_lines:
            print(dftransitions.columns)
            print(dftransitions[[
                'lower', 'upper', 'forbidden', 'A', 'lambda_angstroms', 'flux_factor_upper_pop_lte_3000K']].to_string(
                    index=False))
    print()

    if from_model:
        feions = [2, 3]

        def get_strionfracs(atomic_number, ionstages):
            est_ionfracs = [
                estimators['populations'][(atomic_number, ionstage)] / estimators['populations'][atomic_number]
                for ionstage in ionstages]
            ionfracs_str = ' '.join([f'{pop:6.0e}' if pop < 0.01 else f'{pop:6.2f}' for pop in est_ionfracs])
            strions = ' '.join(
                [f'{at.get_elsymbol(atomic_number)}{at.roman_numerals[ionstage]}'.rjust(6) for ionstage in feions])
            return strions, ionfracs_str

        strfeions, est_fe_ionfracs_str = get_strionfracs(26, [2, 3])

        strniions, est_ni_ionfracs_str = get_strionfracs(28, [2, 3])

        print(f'                     Fe II 7155             Ni II 7378  {strfeions}   /  {strniions}'
              '      T_e    Fe III/II       Ni III/II')

        print(f'{velocity:5.0f} km/s({modelgridindex})      {fe2depcoeff:5.2f}                   '
              f'{ni2depcoeff:.2f}        '
              f'{est_fe_ionfracs_str}   /  {est_ni_ionfracs_str}      {Te:.0f}    '
              f"{estimators['populations'][(26, 3)] / estimators['populations'][(26, 2)]:.2f}          "
              f"{estimators['populations'][(28, 3)] / estimators['populations'][(28, 2)]:5.2f}")

    if from_model:
        outputfilename = str(args.outputfile).format(cell=modelgridindex, timestep=timestep, time_days=time_days)
    else:
        outputfilename = 'plottransitions.pdf'

    make_plot(xvalues, yvalues, temperature_list, vardict, ionlist,
              ionpopdict, args.xmin, args.xmax, figure_title, outputfilename)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



#!/usr/bin/env python3
import argparse
import math
import os
import sys

import matplotlib.pyplot as plt
from artistools import CustomArgHelpFormatter


def addargs(parser):
    parser.add_argument('-inputfolder', action='store', default='.',
                        help='Path to folder with 3D files')

    parser.add_argument('-axis', action='store', dest='chosenaxis', default='x', choices=['x', 'y', 'z'],
                        help='Slice axis (x, y, or z)')

    parser.add_argument('-outputfolder', action='store', default='1dslice',
                        help='Path to folder in which to store 1D output files')

    parser.add_argument('-opdf', action='store', dest='pdfoutputfile',
                        help='Path/filename for PDF plot.')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=CustomArgHelpFormatter,
            description='Covert abundances.txt and model.txt from 3D to a one dimensional slice.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args()

    if not os.path.exists(args.outputfolder):
        os.makedirs(args.outputfolder)
    else:
        if os.path.exists(os.path.join(args.outputfolder, 'model.txt')):
            print('ABORT: model.txt already exists')
            sys.exit()
        elif os.path.exists(os.path.join(args.outputfolder, 'abundances.txt')):
            print('ABORT: abundances.txt already exists')
            sys.exit()

    dict3dcellidto1dcellid, xlist, ylists = slice_3dmodel(args.inputfolder, args.outputfolder, args.chosenaxis)

    slice_abundance_file(args.inputfolder, args.outputfolder, dict3dcellidto1dcellid)

    if args.pdfoutputfile:
        make_plot(xlist, ylists, args.pdfoutputfile)


def slice_3dmodel(inputfolder, outputfolder, chosenaxis):
    xlist = []
    ylists = [[], [], []]
    listout = []
    dict3dcellidto1dcellid = {}
    outcellid = 0
    with open(os.path.join(inputfolder, 'model.txt'), 'r') as fmodelin:
        fmodelin.readline()  # npts_model3d
        t_model = fmodelin.readline()  # days
        fmodelin.readline()  # v_max in [cm/s]

        while True:
            # two lines making up a model grid cell
            block = fmodelin.readline(), fmodelin.readline()

            if not block[0] or not block[1]:
                break

            cell = {}
            blocksplit = block[0].split(), block[1].split()
            if len(blocksplit[0]) == 5:
                (cell['cellid'], cell['pos_x_min'], cell['pos_y_min'], cell['pos_z_min'], cell['rho']) = blocksplit[0]
            else:
                print("Wrong line size")
                sys.exit()

            if len(blocksplit[1]) == 5:
                (cell['ffe'], cell['f56ni'], cell['fco'], cell['f52fe'], cell['f48cr']) = map(float, blocksplit[1])
            else:
                print("Wrong line size")
                sys.exit()

            if cell['pos_x_min'] != "0.0000000" and (chosenaxis != 'x' or float(cell['pos_x_min']) < 0.):
                pass
            elif cell['pos_y_min'] != "0.0000000" and (chosenaxis != 'y' or float(cell['pos_y_min']) < 0.):
                pass
            elif cell['pos_z_min'] != "0.0000000" and (chosenaxis != 'z' or float(cell['pos_z_min']) < 0.):
                pass
            else:
                outcellid += 1
                dict3dcellidto1dcellid[int(cell['cellid'])] = outcellid
                append_cell_to_output(cell, outcellid, t_model, listout, xlist, ylists)
                print(f"Cell {outcellid:4d} input1: {block[0].rstrip()}")
                print(f"Cell {outcellid:4d} input2: {block[1].rstrip()}")
                print(f"Cell {outcellid:4d} output: {listout[-1]}")

    with open(os.path.join(outputfolder, 'model.txt'), 'w') as fmodelout:
        fmodelout.write(f"{outcellid:7d}\n")
        fmodelout.write(t_model)
        for line in listout:
            fmodelout.write(line + "\n")

    return dict3dcellidto1dcellid, xlist, ylists


def slice_abundance_file(inputfolder, outputfolder, dict3dcellidto1dcellid):
    with open(os.path.join(inputfolder, 'abundances.txt'), 'r') as fabundancesin, \
            open(os.path.join(outputfolder, 'abundances.txt'), 'w') as fabundancesout:
        currentblock = []
        keepcurrentblock = False
        for line in fabundancesin:
            linesplit = line.split()

            if len(currentblock) + len(linesplit) >= 30:
                if keepcurrentblock:
                    fabundancesout.write("  ".join(currentblock) + "\n")
                currentblock = []
                keepcurrentblock = False

            if not currentblock:
                currentblock = linesplit
                if int(linesplit[0]) in dict3dcellidto1dcellid:
                    outcellid = dict3dcellidto1dcellid[int(linesplit[0])]
                    currentblock[0] = f"{outcellid:6d}"
                    keepcurrentblock = True
            else:
                currentblock.append(linesplit)

    if keepcurrentblock:
        print("WARNING: unfinished block")


def append_cell_to_output(cell, outcellid, t_model, listout, xlist, ylists):
    dist = math.sqrt(float(cell['pos_x_min']) ** 2 + float(cell['pos_y_min']) ** 2 + float(cell['pos_z_min']) ** 2)
    velocity = float(dist) / float(t_model) / 86400. / 1.e5

    listout.append(
        f"{outcellid:6d}  {velocity:8.2f}  {math.log10(max(float(cell['rho']), 1e-100)):8.5f}  "
        f"{cell['ffe']:.5f}  {cell['f56ni']:.5f}  {cell['fco']:.5f}  {cell['f52fe']:.5f}  {cell['f48cr']:.5f}")

    xlist.append(velocity)
    ylists[0].append(cell['rho'])
    ylists[1].append(cell['f56ni'])
    ylists[2].append(cell['fco'])


def make_plot(xlist, ylists, pdfoutputfile):
    fig, axis = plt.subplots(
        nrows=1, ncols=1, sharey=True, figsize=(6, 4), tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})
    axis.set_xlabel(r'v (km/s)')
    axis.set_ylabel(r'Density (g/cm$^3$) or mass fraction')
    ylabels = [r'$\rho$', 'fNi56', 'fCo']
    for ylist, ylabel in zip(ylists, ylabels):
        axis.plot(xlist, ylist, linewidth=1.5, label=ylabel)
    axis.set_yscale("log", nonposy='clip')
    axis.legend(loc='best', handlelength=2, frameon=False,
                numpoints=1, prop={'size': 10})
    fig.savefig(pdfoutputfile, format='pdf')
    print(f'Saved {pdfoutputfile}')
    plt.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import hashlib
import math
import numpy as np
import os
import os.path
import pandas as pd
import pytest
from pathlib import Path

import artistools as at

modelpath = at.config['path_testartismodel']
outputpath = at.config['path_testoutput']
at.enable_diskcache = False


def test_commands():
    import importlib
    # ensure that the commands are pointing to valid submodule.function() targets
    for command, (submodulename, funcname) in sorted(at.commandlist.items()):
        submodule = importlib.import_module(submodulename, package='artistools')
        assert hasattr(submodule, funcname)


def test_timestep_times():
    timestartarray = at.get_timestep_times_float(modelpath, loc='start')
    timedeltarray = at.get_timestep_times_float(modelpath, loc='delta')
    timemidarray = at.get_timestep_times_float(modelpath, loc='mid')
    assert len(timestartarray) == 100
    assert math.isclose(float(timemidarray[0]), 250.421, abs_tol=1e-3)
    assert math.isclose(float(timemidarray[-1]), 349.412, abs_tol=1e-3)

    assert all([tstart < tmid < (tstart + tdelta)
                for tstart, tdelta, tmid in zip(timestartarray, timedeltarray, timemidarray)])


def test_deposition():
    at.deposition.main(argsraw=[], modelpath=modelpath)


def test_estimator_snapshot():
    at.estimators.main(argsraw=[], modelpath=modelpath, outputfile=outputpath, timedays=300)


def test_estimator_timeevolution():
    at.estimators.main(argsraw=[], modelpath=modelpath, outputfile=outputpath, modelgridindex=0, x='time')


def test_get_inputparams():
    inputparams = at.get_inputparams(modelpath)
    dicthash = hashlib.sha256(str(sorted(inputparams.items())).encode('utf-8')).hexdigest()
    assert dicthash == 'ce7d04d6944207673a105cba8d2430055d0b53b7f3e92db3964d2dca285a3adb'


def test_get_levels():
    at.atomic.get_levels(modelpath, get_transitions=True, get_photoionisations=True)


def test_get_modeldata():
    # expect a 3D model but read 1D
    with pytest.raises(Exception):
        dfmodel, t_model_init_days, vmax_cmps = at.inputmodel.get_modeldata(
            modelpath, get_abundances=True, dimensions=3)

    dfmodel, t_model_init_days, vmax_cmps = at.inputmodel.get_modeldata(modelpath, get_abundances=True)
    assert np.isclose(t_model_init_days, 0.00115740740741, rtol=0.0001)
    assert np.isclose(vmax_cmps, 800000000., rtol=0.0001)

    # assert (
    #     hashlib.sha256(pd.util.hash_pandas_object(dfmodel, index=True).values).hexdigest() ==
    #     '40a02dfa933f6b28671d42f3cf69a182955a5a89dc93bbcd22c894192375fe9b')


def test_lightcurve():
    at.lightcurve.main(argsraw=[], modelpath=modelpath, outputfile=outputpath)


def test_lightcurve_frompackets():
    at.lightcurve.main(argsraw=[], modelpath=modelpath, frompackets=True,
                       outputfile=os.path.join(outputpath, 'lightcurve_from_packets.pdf'))


def test_band_lightcurve_plot():
    at.lightcurve.main(argsraw=[], modelpath=modelpath, filter=['B'], outputfile=outputpath)


def test_band_lightcurve_subplots():
    at.lightcurve.main(argsraw=[], modelpath=modelpath, filter=['bol', 'B'], outputfile=outputpath)


def test_colour_evolution_plot():
    at.lightcurve.main(argsraw=[], modelpath=modelpath, colour_evolution=['B-V'], outputfile=outputpath)


def test_colour_evolution_subplots():
    at.lightcurve.main(argsraw=[], modelpath=modelpath, colour_evolution=['U-B', 'B-V'], outputfile=outputpath)


def test_macroatom():
    at.macroatom.main(argsraw=[], modelpath=modelpath, outputfile=outputpath, timestep=10)


def test_menu():
    at.main(argsraw=[])
    at.showtimesteptimes(modelpath=modelpath)


def test_nltepops():
    # at.nltepops.main(modelpath=modelpath, outputfile=outputpath, timedays=300),
    #                    **benchargs)
    at.nltepops.main(argsraw=[], modelpath=modelpath, outputfile=outputpath, timestep=40)


def test_nonthermal():
    at.nonthermal.main(argsraw=[], modelpath=modelpath, outputfile=outputpath, timestep=70)


def test_radfield():
    at.radfield.main(argsraw=[], modelpath=modelpath, modelgridindex=0, outputfile=outputpath)


def test_get_ionrecombratecalibration():
    at.atomic.get_ionrecombratecalibration(modelpath=modelpath)


def test_spencerfano():
    at.nonthermal.solvespencerfanocmd.main(
        argsraw=[], modelpath=modelpath, timedays=300, makeplot=True, npts=200,
        noexcitation=True, outputfile=outputpath)


def test_transitions():
    at.transitions.main(argsraw=[], modelpath=modelpath, outputfile=outputpath, timedays=300)



#!/usr/bin/env python3

# adapted from Fortran maptogrid.f90 and kernelmodule.f90
# original Fortran code by Andreas Bauswein

import argparse
import math
import numpy as np
import pandas as pd

from pathlib import Path
from artistools import CustomArgHelpFormatter

itable = 40000  # wie fein Kernelfkt interpoliert wird
itab = itable + 5


wij = np.zeros(itab + 1)

#
# --maximum interaction length and step size
#
v2max = 4.
dvtable = v2max / itable
i1 = int(1. // dvtable)


igphi = 0
#
# --normalisation constant
#
cnormk = 1. / math.pi
# --build tables
#
#  a) v less than 1
#
if (igphi == 1):
    for i in range(1, i1 + 1):
        v2 = i * dvtable
        v = math.sqrt(v2)
        v3 = v * v2
        v4 = v * v3
        sum = 1. - 1.5 * v2 + 0.75 * v3
        wij[i] = cnormk * sum
else:
    for i in range(1, i1 + 1):
        v2 = i * dvtable
        v = math.sqrt(v2)
        v3 = v * v2
        sum = 1. - 1.5 * v2 + 0.75 * v3
        wij[i] = cnormk * sum

#
#  b) v greater than 1
#
if (igphi == 1):
    for i in range(i1 + 1, itable + 1):
        v2 = i * dvtable
        v = math.sqrt(v2)
        dif2 = 2. - v
        sum = 0.25 * dif2 * dif2 * dif2
        wij[i] = cnormk * sum
else:
    for i in range(i1 + 1, itable + 1):
        v2 = i * dvtable
        v = math.sqrt(v2)
        dif2 = 2. - v
        sum = 0.25 * dif2 * dif2 * dif2
        wij[i] = cnormk * sum


def kernelvals2(rij2, hmean):  # ist schnell berechnet aber keine Gradienten
    hmean21 = 1. / (hmean * hmean)
    hmean31 = hmean21 / hmean
    v2 = rij2 * hmean21
    index = math.floor(v2 / dvtable)
    dxx = v2 - index * dvtable
    index1 = index + 1
    dwdx = (wij[index1] - wij[index]) / dvtable
    wtij = (wij[index] + dwdx * dxx) * hmean31
    return wtij


def maptogrid(ejectasnapshotpath, outputpath):
    if not ejectasnapshotpath.is_file():
        print(f'{ejectasnapshotpath} not found')
        return

    snapshot_columns = [
        'id', 'h', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'vstx', 'vsty', 'vstz', 'u',
        'psi', 'alpha', 'pmass', 'rho', 'p', 'rst', 'tau', 'av', 'ye', 'temp',
        'prev_rho(i)', 'ynue(i)', 'yanue(i)', 'enuetrap(i)', 'eanuetrap(i)',
        'enuxtrap(i)', 'iwasequil(i, 1)', 'iwasequil(i, 2)', 'iwasequil(i, 3)']
    snapshot_columns_used = ['id', 'h', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'pmass', 'rho', 'p', 'rst', 'ye']
    dfsnapshot = pd.read_csv(
        ejectasnapshotpath, names=snapshot_columns,
        delim_whitespace=True, usecols=snapshot_columns_used)

    print(dfsnapshot)

    assert len(dfsnapshot.columns) == len(snapshot_columns_used)

    npart = len(dfsnapshot)
    print("number of particles", npart)

    fpartanalysis = open(Path(outputpath, 'ejectapartanalysis.dat'), mode='w')

    totmass = 0.0
    rmax = 0.0
    rmean = 0.0
    hmean = 0.0
    hmin = 100000.
    vratiomean = 0.

    # Propagate particles to dtextra using velocities
    dtextra = 0.5  # in seconds ---  dtextra = 0.0 # for no extrapolation

    dtextra = dtextra / 4.926e-6  # convert to geom units.

    particleid = dfsnapshot.id.values
    x = dfsnapshot.x.values
    y = dfsnapshot.y.values
    z = dfsnapshot.z.values
    h = dfsnapshot.h.values
    vx = dfsnapshot.vx.values
    vy = dfsnapshot.vy.values
    vz = dfsnapshot.vz.values
    pmass = dfsnapshot.pmass.values
    rst = dfsnapshot.rst.values
    Ye = dfsnapshot.ye.values

    for n in range(npart):
        totmass = totmass + pmass[n]

        dis = math.sqrt(x[n] ** 2 + y[n] ** 2 + z[n] ** 2)  # original dis

        x[n] += vx[n] * dtextra
        y[n] += vy[n] * dtextra
        z[n] += vz[n] * dtextra

        # actually we should also extrapolate smoothing length h unless we disrgard it below

        # extrapolate h such that ratio betwen dis and h remains constant
        h[n] = h[n] / dis * math.sqrt(x[n] ** 2 + y[n] ** 2 + z[n] ** 2)

        dis = math.sqrt(x[n] ** 2 + y[n] ** 2 + z[n] ** 2)  # possibly new distance

        rmean = rmean + dis

        rmax = max(rmax, dis)

        hmean = hmean + h[n]

        hmin = min(hmin, h[n])

        vtot = math.sqrt(vx[n] ** 2 + vy[n] ** 2 + vz[n] ** 2)

        vrad = (vx[n] * x[n] + vy[n] * y[n] + vz[n] * z[n]) / dis  # radial velocity

        if (vtot > vrad):
            vperp = math.sqrt(vtot * vtot - vrad * vrad)  # velicty perpendicular
        else:
            vperp = 0.0  # if we rxtrapolate roundoff error can lead to Nan, ly?

        vratiomean = vratiomean + vperp / vrad

        # output some ejecta properties in file

        fpartanalysis.write(f'{dis} {h[n]} {h[n] / dis} {vrad} {vperp} {vtot}\n')

    rmean = rmean / npart

    hmean = hmean / npart

    vratiomean = vratiomean / npart

    print("total mass of sph particle, max, mean distance", totmass, rmax, rmean)
    print("smoothing length min, mean", hmin, hmean)
    print("ratio between vrad and vperp mean", vratiomean)

    # check maybe cm and correct by shifting

    # ...

    # set up grid

    ngrid = 50

    x0 = - 0.6 * rmax  # 90% is hand waving - choose #

    # x0 = - rmean

    dx = 2. * abs(x0) / (ngrid)  # -1 to be symmetric, right?

    y0 = x0
    z0 = x0
    dy = dx
    dz = dx

    grho = np.zeros((ngrid + 1, ngrid + 1, ngrid + 1))
    norm = np.zeros((ngrid + 1, ngrid + 1, ngrid + 1))
    gye = np.zeros((ngrid + 1, ngrid + 1, ngrid + 1))
    gparticlecounter = np.zeros((ngrid + 1, ngrid + 1, ngrid + 1), dtype=int)
    particlecontribs = {}

    print('grid properties', x0, dx, x0 + dx * (ngrid - 1))

    arrgx = x0 + dx * (np.arange(ngrid + 1) - 1)
    arrgy = arrgx
    arrgz = arrgx

    for n in range(npart):

        maxdist = 2. * h[n]
        maxdist2 = maxdist ** 2

        ilow = max(math.floor((x[n] - maxdist - x0) / dx), 1)
        ihigh = min(math.ceil((x[n] + maxdist - x0) / dx), ngrid)
        jlow = max(math.floor((y[n] - maxdist - y0) / dy), 1)
        jhigh = min(math.ceil((y[n] + maxdist - y0) / dy), ngrid)
        klow = max(math.floor((z[n] - maxdist - z0) / dz), 1)
        khigh = min(math.ceil((z[n] + maxdist - z0) / dz), ngrid)

        # check some min max

        # ... kernel reweighting ?

        searchcoords = [
            (i, j, k, (arrgx[i] - x[n]) ** 2 + (arrgy[j] - y[n]) ** 2 + (arrgz[k] - z[n]) ** 2)
            for i in range(ilow, ihigh + 1)
            for j in range(jlow, jhigh + 1)
            for k in range(klow, khigh + 1)
        ]

        for i, j, k, dis2 in searchcoords:
            # -- change h by hand --------- we could do these particle thinsg also further up

            # option 1 minimum that no particle is lost

            # option 2 increase smoothing everywhere, i.e. less holes but also less strcuture

            # option 3 increase smoothing beyond some distance

            # options can be combined, i.e. option 1 alone fills the hole in the center
            # (which we could also replace by later ejecta)

            # h[n] = max(h[n],1.5*dx) # option 1

            # h[n] = max(h[n],0.25*dis) #  option 2

            # dis = sqrt(x[n]*x[n]+y[n]*y[n]+z[n]*z[n])

            # if (dis>1.5*rmean) h[n]=max(h[n],0.4*dis) # option 3

            # -------------------------------

            # or via neighbors  - not yet implemented

            if dis2 <= maxdist2:

                wtij = kernelvals2(dis2, h[n])

                grho[i, j, k] += pmass[n] * wtij
                # grho[i, j, k] = grho[i, j, k] + pmass[n] / particle.rst * wtij

                particle_contrib = pmass[n] / rst[n] * wtij
                particlecontribs[(n, i, j, k)] = particle_contrib

                # gye[i, j, k] += pmass[n] * particle.Ye * wtij
                gye[i, j, k] += particle_contrib * Ye[n]

                # norm[i, j, k] += pmass[n] * wtij
                norm[i, j, k] += particle_contrib

                # Counts particles contributing to grid cell
                gparticlecounter[i, j, k] += 1

    with np.errstate(divide='ignore', invalid='ignore'):
        gye = np.divide(gye, norm)
        # for i in range(1, ngrid + 1):
        #     for j in range(1, ngrid + 1):
        #         for k in range(1, ngrid + 1):
        #             gye[i, j, k] = gye[i, j, k] / norm[i, j, k]
        #             # if not norm[i, j, k] > 0:
        #             #     print(i, j, k, norm[i, j, k], gye[i, j, k] )

        with open(Path(outputpath, 'gridcontributions.txt'), 'w') as fcontribs:
            fcontribs.write('particleid cellindex frac_of_cellmass\n')
            for (n, i, j, k), old_value in particlecontribs.items():
                # particle_contribs[n, i, j, k] = old_value / norm[i, j, k]
                gridindex = ((k - 1) * ngrid + (j - 1)) * ngrid + (i - 1) + 1
                fcontribs.write(f'{particleid[n]} {gridindex} {old_value / norm[i, j, k]}\n')
    # check some stuff on the grid

    gmass = 0.0
    nzero = 0
    nzerocentral = 0
    gmass = np.sum(grho) * dx * dy * dz
    # nzero = np.count_nonzero(grho[1:][1:][1:] < 1.e-20)

    for i in range(1, ngrid + 1):
        gx = x0 + dx * (i - 1)
        for j in range(1, ngrid + 1):
            gy = y0 + dy * (j - 1)
            for k in range(1, ngrid + 1):

                # how many cells with rho=0?

                if grho[i, j, k] < 1.e-20:
                    nzero = nzero + 1

                gz = z0 + dz * (k - 1)

                dis = math.sqrt(gx * gx + gy * gy + gz * gz)

                if (grho[i, j, k] < 1.e-20 and dis < rmean):
                    nzerocentral = nzerocentral + 1

    print("mass on grid and particles", gmass, totmass)

    print("number of cells with zero rho, total num of cells, fraction of cells w rho=0",
          nzero, ngrid**3, (nzero) / (ngrid**3))

    print("number of central cells (dis<rmean) with zero rho, ratio",
          nzerocentral, (nzerocentral) / (4. * 3.14 / 3. * rmean**3 / (dx * dy * dz)))

    print("probably we want to choose grid size, i.e. x0, as compromise between mapped mass and rho=0 cells")

    # output grid - adapt as you need output

    with open(Path(outputpath, 'grid.dat'), 'w') as fgrid:

        fgrid.write(f'{ngrid**3} # ngrid\n')
        fgrid.write(f'{dtextra} # extra time after explosion simulation ended (in geom units)\n')
        fgrid.write(f'{x0} # xmax\n')
        fgrid.write(' gridindex    pos_x_min    pos_y_min    pos_z_min    rho    cellYe    tracercount\n')
        gridindex = 1
        for k in range(1, ngrid + 1):
            gz = z0 + dz * (k - 1)
            for j in range(1, ngrid + 1):
                gy = y0 + dy * (j - 1)
                for i in range(1, ngrid + 1):
                    fgrid.write(
                        f'{gridindex:8d} {x0+dx*(i-1)} {gy} {gz} '
                        f'{grho[i,j,k]} {gye[i,j,k]} {gparticlecounter[i,j,k]}\n')
                    # gridindex2 = ((k - 1) * ngrid + (j - 1)) * ngrid + (i - 1) + 1

                    gridindex = gridindex + 1


def addargs(parser):
    parser.add_argument('-inputpath', '-i',
                        default='.',
                        help='Path to ejectasnapshot')
    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path for output files')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=CustomArgHelpFormatter,
            description='Map tracer particle trajectories to a Cartesian grid.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    ejectasnapshotpath = Path(args.inputpath, 'ejectasnapshot.dat')

    maptogrid(ejectasnapshotpath=ejectasnapshotpath, outputpath=args.outputpath)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Functions for reading and processing estimator files.

Examples are temperatures, populations, and heating/cooling rates.
"""
# import math
import math
import multiprocessing
import sys
from collections import namedtuple
from functools import lru_cache, partial, reduce
# from itertools import chain
from pathlib import Path

import pandas as pd

import artistools as at
import artistools.nltepops


variableunits = {
    'time': 'days',
    'gamma_NT': '/s',
    'gamma_R_bfest': '/s',
    'TR': 'K',
    'Te': 'K',
    'TJ': 'K',
    'nne': 'e-/cm3',
    'heating': 'erg/s/cm3',
    'heating_dep/total_dep': 'Ratio',
    'cooling': 'erg/s/cm3',
    'velocity': 'km/s',
    'velocity_outer': 'km/s',
}

variablelongunits = {
    'heating_dep/total_dep': '',
    'TR': 'Temperature [K]',
    'Te': 'Temperature [K]',
    'TJ': 'Temperature [K]',
}

dictlabelreplacements = {
    'lognne': 'Log nne',
    'Te': 'T$_e$',
    'TR': 'T$_R$',
    'gamma_NT': r'$\Gamma_{\rm non-thermal}$ [s$^{-1}$]',
    'gamma_R_bfest': r'$\Gamma_{\rm phot}$ [s$^{-1}$]',
    'heating_dep/total_dep': 'Heating fraction',
}


def apply_filters(xlist, ylist, args):
    filterfunc = at.get_filterfunc(args)

    if filterfunc is not None:
        ylist = filterfunc(ylist)

    return xlist, ylist


def get_ionrecombrates_fromfile(filename):
    """WARNING: copy pasted from artis-atomic! replace with a package import soon ionstage is the lower ion stage."""
    print(f'Reading {filename}')

    header_row = []
    with open(filename, 'r') as filein:
        while True:
            line = filein.readline()
            if line.strip().startswith('TOTAL RECOMBINATION RATE'):
                line = filein.readline()
                line = filein.readline()
                header_row = filein.readline().strip().replace(' n)', '-n)').split()
                break

        if not header_row:
            print("ERROR: no header found")
            sys.exit()

        index_logt = header_row.index('log(T)')
        index_low_n = header_row.index('RRC(low-n)')
        index_tot = header_row.index('RRC(total)')

        recomb_tuple = namedtuple("recomb_tuple", ['logT', 'RRC_low_n', 'RRC_total'])
        records = []
        for line in filein:
            row = line.split()
            if row:
                if len(row) != len(header_row):
                    print('Row contains wrong number of items for header:')
                    print(header_row)
                    print(row)
                    sys.exit()
                records.append(recomb_tuple(
                    *[float(row[index]) for index in [index_logt, index_low_n, index_tot]]))

    dfrecombrates = pd.DataFrame.from_records(records, columns=recomb_tuple._fields)
    return dfrecombrates


def get_units_string(variable):
    if variable in variableunits:
        return f' [{variableunits[variable]}]'
    elif variable.split('_')[0] in variableunits:
        return f' [{variableunits[variable.split("_")[0]]}]'
    return ''


def parse_estimfile(estfilepath, modelpath, get_ion_values=True, get_heatingcooling=True):
    """Generate timestep, modelgridindex, dict from estimator file."""
    # itstep = at.get_inputparams(modelpath)['itstep']

    with at.zopen(estfilepath, 'rt') as estimfile:
        timestep = -1
        modelgridindex = -1
        estimblock = {}
        for line in estimfile:
            row = line.split()
            if not row:
                continue

            if row[0] == 'timestep':
                # yield the previous block before starting a new one
                if timestep >= 0 and modelgridindex >= 0:
                    yield timestep, modelgridindex, estimblock

                timestep = int(row[1])
                # if timestep > itstep:
                #     print(f"Dropping estimator data from timestep {timestep} and later (> itstep {itstep})")
                #     # itstep in input.txt is updated by ARTIS at every timestep, so the data beyond here
                #     # could be half-written to disk and cause parsing errors
                #     return

                modelgridindex = int(row[3])
                # print(f'Timestep {timestep} cell {modelgridindex}')

                estimblock = {}
                emptycell = (row[4] == 'EMPTYCELL')
                estimblock['emptycell'] = emptycell
                if not emptycell:
                    # will be TR, Te, W, TJ, nne
                    for variablename, value in zip(row[4::2], row[5::2]):
                        estimblock[variablename] = float(value)
                    estimblock['lognne'] = math.log10(estimblock['nne']) if estimblock['nne'] > 0 else float('-inf')

            elif row[1].startswith('Z=') and get_ion_values:
                variablename = row[0]
                if row[1].endswith('='):
                    atomic_number = int(row[2])
                    startindex = 3
                else:
                    atomic_number = int(row[1].split('=')[1])
                    startindex = 2

                estimblock.setdefault(variablename,  {})

                for ion_stage_str, value in zip(row[startindex::2], row[startindex + 1::2]):
                    if ion_stage_str.strip() in ['SUM:', '(or']:
                        continue

                    try:
                        ion_stage = int(ion_stage_str.rstrip(':'))
                    except ValueError:
                        if variablename == 'populations' and ion_stage_str.startswith(at.get_elsymbol(atomic_number)):
                            estimblock[variablename][ion_stage_str.rstrip(':')] = float(value)
                        else:
                            print(ion_stage_str, at.get_elsymbol(atomic_number))
                            print(f'Cannot parse row: {row}')
                        continue

                    value_thision = float(value.rstrip(','))

                    estimblock[variablename][(atomic_number, ion_stage)] = value_thision

                    if variablename in ['Alpha_R*nne', 'AlphaR*nne']:
                        estimblock.setdefault('Alpha_R', {})
                        estimblock['Alpha_R'][(atomic_number, ion_stage)] = value_thision / estimblock['nne']

                    else:  # variablename == 'populations':

                        # contribute the ion population to the element population
                        estimblock[variablename].setdefault(atomic_number, 0.)
                        estimblock[variablename][atomic_number] += value_thision

                if variablename == 'populations':
                    # contribute the element population to the total population
                    estimblock['populations'].setdefault('total', 0.)
                    estimblock['populations']['total'] += estimblock['populations'][atomic_number]
                    estimblock['nntot'] = estimblock['populations']['total']

            elif row[0] == 'heating:' and get_heatingcooling:
                for heatingtype, value in zip(row[1::2], row[2::2]):
                    key = 'heating_' + heatingtype if not heatingtype.startswith('heating_') else heatingtype
                    estimblock[key] = float(value)

                if 'heating_gamma/gamma_dep' in estimblock and estimblock['heating_gamma/gamma_dep'] > 0:
                    estimblock['gamma_dep'] = (
                        estimblock['heating_gamma'] /
                        estimblock['heating_gamma/gamma_dep'])
                elif 'heating_dep/total_dep' in estimblock and estimblock['heating_dep/total_dep'] > 0:
                    estimblock['total_dep'] = (
                        estimblock['heating_dep'] /
                        estimblock['heating_dep/total_dep'])

            elif row[0] == 'cooling:' and get_heatingcooling:
                for coolingtype, value in zip(row[1::2], row[2::2]):
                    estimblock['cooling_' + coolingtype] = float(value)

    # reached the end of file
    if timestep >= 0 and modelgridindex >= 0:
        yield timestep, modelgridindex, estimblock


@at.diskcache(ignorekwargs=['printfilename'], quiet=False, funcdepends=parse_estimfile, savezipped=True)
def read_estimators_from_file(folderpath, modelpath, arr_velocity_outer, mpirank, printfilename=False,
                              get_ion_values=True, get_heatingcooling=True):

    estimators_thisfile = {}
    estimfilename = f'estimators_{mpirank:04d}.out'
    estfilepath = Path(folderpath, estimfilename)
    if not estfilepath.is_file():
        estfilepath = Path(folderpath, estimfilename + '.gz')
        if not estfilepath.is_file():
            estfilepath = Path(folderpath, estimfilename + '.xz')
            if not estfilepath.is_file():
                # not worth printing and error, because ranks with no cells to update do not produce an estimator file
                # print(f'Warning: Could not find {estfilepath.relative_to(modelpath.parent)}')
                return {}

    if printfilename:
        filesize = Path(estfilepath).stat().st_size / 1024 / 1024
        print(f'Reading {estfilepath.relative_to(modelpath.parent)} ({filesize:.2f} MiB)')

    for fileblock_timestep, fileblock_modelgridindex, file_estimblock in parse_estimfile(
            estfilepath, modelpath, get_ion_values=get_ion_values, get_heatingcooling=get_heatingcooling):

        if arr_velocity_outer is not None:
            file_estimblock['velocity_outer'] = arr_velocity_outer[fileblock_modelgridindex]
            file_estimblock['velocity'] = file_estimblock['velocity_outer']

        estimators_thisfile[(fileblock_timestep, fileblock_modelgridindex)] = file_estimblock

    return estimators_thisfile


@lru_cache(maxsize=16)
@at.diskcache(savezipped=True, funcdepends=[read_estimators_from_file, parse_estimfile])
def read_estimators(modelpath, modelgridindex=None, timestep=None, get_ion_values=True, get_heatingcooling=True):
    """Read estimator files into a nested dictionary structure.

    Speed it up by only retrieving estimators for a particular timestep(s) or modelgrid cells.
    """

    if modelgridindex is None:
        match_modelgridindex = []
    elif hasattr(modelgridindex, '__iter__'):
        match_modelgridindex = tuple(modelgridindex)
    else:
        match_modelgridindex = (modelgridindex,)

    if -1 in match_modelgridindex:
        match_modelgridindex = []

    if timestep is None:
        match_timestep = []
    else:
        match_timestep = tuple(timestep) if hasattr(timestep, '__iter__') else (timestep,)

    if not modelpath.exists() and modelpath.parts[0] == 'codecomparison':
        return artistools.codecomparison.read_reference_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)

    # print(f" matching cells {match_modelgridindex} and timesteps {match_timestep}")

    modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
    if 'velocity_outer' in modeldata.columns:
        arr_velocity_outer = tuple(list([float(v) for v in modeldata['velocity_outer'].values]))
    else:
        arr_velocity_outer = None

    mpiranklist = at.get_mpiranklist(modelpath, modelgridindex=match_modelgridindex, only_ranks_withgridcells=True)

    printfilename = len(mpiranklist) < 10

    estimators = {}
    for folderpath in at.get_runfolders(modelpath, timesteps=match_timestep):
        print(f'Reading {len(list(mpiranklist))} estimator files in {folderpath.relative_to(Path(modelpath).parent)}')

        processfile = partial(read_estimators_from_file, folderpath, modelpath, arr_velocity_outer,
                              get_ion_values=get_ion_values, get_heatingcooling=get_heatingcooling,
                              printfilename=printfilename)

        if at.num_processes > 1:
            with multiprocessing.Pool(processes=at.num_processes) as pool:
                arr_rankestimators = pool.map(processfile, mpiranklist)
                pool.close()
                pool.join()
                pool.terminate()
        else:
            arr_rankestimators = [processfile(rank) for rank in mpiranklist]

        for mpirank, estimators_thisfile in zip(mpiranklist, arr_rankestimators):
            dupekeys = list(sorted([k for k in estimators_thisfile if k in estimators]))
            for k in dupekeys:
                # dropping the lowest timestep is normal for restarts. Only warn about other cases
                if k[0] != dupekeys[0][0]:
                    filepath = Path(folderpath, f'estimators_{mpirank:04d}.out')
                    print(f'WARNING: Duplicate estimator block for (timestep, mgi) key {k}. '
                          f'Dropping block from {filepath}')

                del estimators_thisfile[k]

            estimators.update(estimators_thisfile)

    return estimators


def get_averaged_estimators(modelpath, estimators, timesteps, modelgridindex, keys, avgadjcells=0):
    """Get the average of estimators[(timestep, modelgridindex)][keys[0]]...[keys[-1]] across timesteps."""
    if isinstance(keys, str):
        keys = [keys]

    # reduce(lambda d, k: d[k], keys, dictionary) returns dictionary[keys[0]][keys[1]]...[keys[-1]]
    # applying all keys in the keys list

    # if single timestep, no averaging needed
    if not hasattr(timesteps, '__iter__'):
        return reduce(lambda d, k: d[k], [(timesteps, modelgridindex)] + keys, estimators)

    firsttimestepvalue = reduce(lambda d, k: d[k], [(timesteps[0], modelgridindex)] + keys, estimators)
    if isinstance(firsttimestepvalue, dict):
        dictout = {k: get_averaged_estimators(modelpath, estimators, timesteps, modelgridindex, keys + [k])
                   for k in firsttimestepvalue.keys()}

        return dictout
    else:
        tdeltas = at.get_timestep_times_float(modelpath, loc='delta')
        valuesum = 0
        tdeltasum = 0
        for timestep, tdelta in zip(timesteps, tdeltas):
            for mgi in range(modelgridindex - avgadjcells, modelgridindex + avgadjcells + 1):
                try:
                    valuesum += reduce(lambda d, k: d[k], [(timestep, mgi)] + keys, estimators) * tdelta
                    tdeltasum += tdelta
                except KeyError:
                    pass
        return valuesum / tdeltasum

    # except KeyError:
    #     if (timestep, modelgridindex) in estimators:
    #         print(f'Unknown x variable: {xvariable} for timestep {timestep} in cell {modelgridindex}')
    #     else:
    #         print(f'No data for cell {modelgridindex} at timestep {timestep}')
    #     print(estimators[(timestep, modelgridindex)])
    #     sys.exit()


def get_averageionisation(populations, atomic_number):
    free_electron_weighted_pop_sum = 0.
    found = False
    popsum = 0
    for key in populations.keys():
        if isinstance(key, tuple) and key[0] == atomic_number:
            found = True
            ion_stage = key[1]
            free_electron_weighted_pop_sum += populations[key] * (ion_stage - 1)
            popsum += populations[key]

    if not found:
        return float('NaN')

    return free_electron_weighted_pop_sum / populations[atomic_number]


def get_averageexcitation(modelpath, modelgridindex, timestep, atomic_number, ion_stage, T_exc):
    import artistools.nltepops
    dfnltepops = at.nltepops.read_files(modelpath, modelgridindex=modelgridindex, timestep=timestep)
    adata = at.atomic.get_levels(modelpath)
    ionlevels = adata.query('Z == @atomic_number and ion_stage == @ion_stage').iloc[0].levels

    energypopsum = 0
    ionpopsum = 0
    if dfnltepops.empty:
        return float('NaN')
    else:
        dfnltepops_ion = dfnltepops.query(
            'modelgridindex==@modelgridindex and timestep==@timestep and Z==@atomic_number & ion_stage==@ion_stage')

        k_b = 8.617333262145179e-05  # eV / K

        ionpopsum = dfnltepops_ion.n_NLTE.sum()
        energypopsum = dfnltepops_ion[dfnltepops_ion.level >= 0].eval(
            '@ionlevels.iloc[level].energy_ev.values * n_NLTE').sum()

        try:
            superlevelrow = dfnltepops_ion[dfnltepops_ion.level < 0].iloc[0]
            levelnumber_sl = dfnltepops_ion.level.max() + 1

            energy_boltzfac_sum = ionlevels.iloc[levelnumber_sl:].eval(
                'energy_ev * g * exp(- energy_ev / @k_b / @T_exc)').sum()

            boltzfac_sum = ionlevels.iloc[levelnumber_sl:].eval('g * exp(- energy_ev / @k_b / @T_exc)').sum()
            # adjust to the actual superlevel population from ARTIS
            energypopsum += energy_boltzfac_sum * superlevelrow.n_NLTE / boltzfac_sum
        except IndexError:
            # no superlevel
            pass

    return energypopsum / ionpopsum



#!/usr/bin/env python3

import math
import gzip
# import multiprocessing
import sys
from pathlib import Path

# import matplotlib.patches as mpatches
import numpy as np
import pandas as pd

# from collections import namedtuple
from functools import lru_cache

import artistools as at

CLIGHT = 2.99792458e10
DAY = 86400

types = {
    10: 'TYPE_GAMMA',
    11: 'TYPE_RPKT',
    20: 'TYPE_NTLEPTON',
    32: 'TYPE_ESCAPE',
}

type_ids = dict((v, k) for k, v in types.items())


@lru_cache(maxsize=16)
def get_column_names_artiscode(modelpath):
    modelpath = Path(modelpath)
    if Path(modelpath, 'artis').is_dir():
        print('detected artis code directory')
        packet_properties = []
        inputfilename = at.firstexisting(
            ['packet_init.cc', 'packet_init.c'], path=(modelpath / 'artis'))
        print(f'found {inputfilename}: getting packet column names from artis code')
        with open(inputfilename) as inputfile:
            packet_print_lines = [line.split(',') for line in inputfile if 'fprintf(packets_file,' in line]
            for line in packet_print_lines:
                for element in line:
                    if 'pkt[i].' in element:
                        packet_properties.append(element)

        for i, element in enumerate(packet_properties):
            packet_properties[i] = element.split('.')[1].split(')')[0]

        columns = packet_properties
        replacements_dict = {
            'type': 'type_id',
            'pos[0]': 'posx', 'pos[1]': 'posy', 'pos[2]': 'posz',
            'dir[0]': 'dirx', 'dir[1]': 'diry', 'dir[2]': 'dirz',
            'escape_type': 'escape_type_id',
            'em_pos[0]': 'em_posx', 'em_pos[1]': 'em_posy', 'em_pos[2]': 'em_posz',
            'absorptiontype': 'absorption_type', 'absorptionfreq': 'absorption_freq',
            'absorptiondir[0]': 'absorptiondirx', 'absorptiondir[1]': 'absorptiondiry',
            'absorptiondir[2]': 'absorptiondirz',
            'stokes[0]': 'stokes1', 'stokes[1]': 'stokes2', 'stokes[2]': 'stokes3',
            'pol_dir[0]': 'pol_dirx', 'pol_dir[1]': 'pol_diry', 'pol_dir[2]': 'pol_dirz'
        }

        for i, column_name in enumerate(columns):
            if column_name in replacements_dict:
                columns[i] = replacements_dict[column_name]

        return columns

    return False


def add_derived_columns(dfpackets, modelpath, colnames, allnonemptymgilist=None):
    cm_to_km = 1e-5
    day_in_s = 86400
    if dfpackets.empty:
        return dfpackets

    colnames = at.makelist(colnames)

    def em_modelgridindex(packet):
        return at.get_mgi_of_velocity_kms(modelpath, packet.emission_velocity * cm_to_km,
                                          mgilist=allnonemptymgilist)

    def emtrue_modelgridindex(packet):
        return at.get_mgi_of_velocity_kms(modelpath, packet.true_emission_velocity * cm_to_km,
                                          mgilist=allnonemptymgilist)

    def em_timestep(packet):
        return at.get_timestep_of_timedays(modelpath, packet.em_time / day_in_s)

    def emtrue_timestep(packet):
        return at.get_timestep_of_timedays(modelpath, packet.trueem_time / day_in_s)

    if 'emission_velocity' in colnames:
        dfpackets.eval(
            "emission_velocity = sqrt(em_posx ** 2 + em_posy ** 2 + em_posz ** 2) / em_time",
            inplace=True)

        dfpackets.eval(
            "em_velx = em_posx / em_time",
            inplace=True)
        dfpackets.eval(
            "em_vely = em_posy / em_time",
            inplace=True)
        dfpackets.eval(
            "em_velz = em_posz / em_time",
            inplace=True)

    if 'em_modelgridindex' in colnames:
        if 'emission_velocity' not in dfpackets.columns:
            dfpackets = add_derived_columns(dfpackets, modelpath, ['emission_velocity'],
                                            allnonemptymgilist=allnonemptymgilist)
        dfpackets['em_modelgridindex'] = dfpackets.apply(em_modelgridindex, axis=1)

    if 'emtrue_modelgridindex' in colnames:
        dfpackets['emtrue_modelgridindex'] = dfpackets.apply(emtrue_modelgridindex, axis=1)

    if 'em_timestep' in colnames:
        dfpackets['em_timestep'] = dfpackets.apply(em_timestep, axis=1)

    return dfpackets


def readfile_text(packetsfile, modelpath=Path('.')):
    column_names = None
    try:
        fpackets = at.zopen(packetsfile, 'rt')

        datastartpos = fpackets.tell()  # will be updated if this was actually the start of a header
        firstline = fpackets.readline()

        if firstline.lstrip().startswith('#'):
            column_names = firstline.lstrip('#').split()
            # get the column count from the first data line to check header matched
            datastartpos = fpackets.tell()
            dataline = fpackets.readline()
            inputcolumncount = len(dataline.split())
        else:
            inputcolumncount = len(firstline.split())

        fpackets.seek(datastartpos)  # go to first data line

    except gzip.BadGzipFile:
        print(f"\nBad Gzip File: {packetsfile}")
        raise gzip.BadGzipFile

    if inputcolumncount < 3:
        print("\nWARNING: packets file has no columns!")
        print(open(packetsfile, "r").readlines())

    usecols_nodata = None  # print a warning for missing columns if the source code columns can't be read
    if column_names:  # found a header in the file
        assert len(column_names) == inputcolumncount
    else:
        column_names = get_column_names_artiscode(modelpath)
        if column_names:  # found them in the artis code files
            assert len(column_names) == inputcolumncount
        else:  # infer from positions
            # new artis added extra columns to the end of this list, but they may be absent in older versions
            # the packets file may have a truncated set of columns, but we assume that they
            # are only truncated, i.e. the columns with the same index have the same meaning
            columns_full = (
                'number',
                'where',
                'type_id',
                'posx', 'posy', 'posz',
                'dirx', 'diry', 'dirz',
                'last_cross',
                'tdecay',
                'e_cmf',
                'e_rf',
                'nu_cmf',
                'nu_rf',
                'escape_type_id',
                'escape_time',
                'scat_count',
                'next_trans',
                'interactions',
                'last_event',
                'emissiontype',
                'trueemissiontype',
                'em_posx', 'em_posy', 'em_posz',
                'absorption_type',
                'absorption_freq',
                'nscatterings',
                'em_time',
                'absorptiondirx',
                'absorptiondiry',
                'absorptiondirz', 'stokes1', 'stokes2', 'stokes3', 'pol_dirx', 'pol_diry',
                'pol_dirz',
                'originated_from_positron',
                'true_emission_velocity',
                'trueem_time',
                'pellet_nucindex',
            )

            assert len(columns_full) >= inputcolumncount
            usecols_nodata = [n for n in columns_full if columns_full.index(n) >= inputcolumncount]
            column_names = columns_full[:inputcolumncount]

    try:
        dfpackets = pd.read_csv(fpackets, delim_whitespace=True, names=column_names, header=None)

    except Exception as ex:
        print(f'Problem with file {packetsfile}')
        print(f'ERROR: {ex}')
        fpackets.close()
        sys.exit(1)

    fpackets.close()

    if usecols_nodata:
        print(f'WARNING: no data in packets file for columns: {usecols_nodata}')
        for col in usecols_nodata:
            dfpackets[col] = float('NaN')

    return dfpackets


@at.diskcache(savezipped=True)
def readfile(packetsfile, type=None, escape_type=None):
    """Read a packet file into a pandas DataFrame."""
    packetsfile = Path(packetsfile)

    if packetsfile.suffixes == ['.out', '.parquet']:
        dfpackets = pd.read_parquet(packetsfile)
    elif packetsfile.suffixes == ['.out', '.feather']:
        dfpackets = pd.read_feather(packetsfile)
    elif packetsfile.suffixes in [['.out'], ['.out', '.gz'], ['.out', '.xz']]:
        dfpackets = readfile_text(packetsfile)
        # dfpackets.to_parquet(at.stripallsuffixes(packetsfile).with_suffix('.out.parquet'),
        #                      compression='brotli', compression_level=99)
    else:
        print('ERROR')
        sys.exit(1)
    filesize = Path(packetsfile).stat().st_size / 1024 / 1024
    print(f'Reading {packetsfile} ({filesize:.1f} MiB)', end='')

    print(f' ({len(dfpackets):.1e} packets', end='')

    if escape_type is not None and escape_type != '' and escape_type != 'ALL':
        assert type is None or type == 'TYPE_ESCAPE'
        dfpackets.query(f'type_id == {type_ids["TYPE_ESCAPE"]} and escape_type_id == {type_ids[escape_type]}',
                        inplace=True)
        print(f', {len(dfpackets)} escaped as {escape_type})')
    elif type is not None and type != 'ALL' and type != '':
        dfpackets.query(f'type_id == {type_ids[type]}', inplace=True)
        print(f', {len(dfpackets)} with type {type})')
    else:
        print(')')

    # dfpackets['type'] = dfpackets['type_id'].map(lambda x: types.get(x, x))
    # dfpackets['escape_type'] = dfpackets['escape_type_id'].map(lambda x: types.get(x, x))

    # # neglect light travel time correction
    # dfpackets.eval("t_arrive_d = escape_time / 86400", inplace=True)

    dfpackets.eval(
        "t_arrive_d = (escape_time - (posx * dirx + posy * diry + posz * dirz) / 29979245800) / 86400", inplace=True)

    return dfpackets


@lru_cache(maxsize=16)
def get_packetsfilepaths(modelpath, maxpacketfiles=None):

    def preferred_alternative(f, files):
        f_nosuffixes = at.stripallsuffixes(f)

        suffix_priority = [['.out', '.gz'], ['.out', '.xz'], ['.out', '.feather'], ['.out', '.parquet']]
        if f.suffixes in suffix_priority:
            startindex = suffix_priority.index(f.suffixes) + 1
        else:
            startindex = 0

        if any(f_nosuffixes.with_suffix(''.join(s)).is_file() for s in suffix_priority[startindex:]):
            return True
        return False

    packetsfiles = sorted(
        list(Path(modelpath).glob('packets00_*.out*')) +
        list(Path(modelpath, 'packets').glob('packets00_*.out*')))

    # strip out duplicates in the case that some are stored as binary and some are text files
    packetsfiles = [f for f in packetsfiles if not preferred_alternative(f, packetsfiles)]

    if maxpacketfiles is not None and maxpacketfiles > 0 and len(packetsfiles) > maxpacketfiles:
        print(f'Using only the first {maxpacketfiles} of {len(packetsfiles)} packets files')
        packetsfiles = packetsfiles[:maxpacketfiles]

    return packetsfiles


def get_escaping_packet_angle_bin(modelpath, dfpackets):
    MABINS = 100

    syn_dir = at.get_syn_dir(modelpath)

    angle_number = np.zeros(len(dfpackets))
    i = 0
    for pkt_index, _ in dfpackets.iterrows():
        pkt_dir = [dfpackets['dirx'][pkt_index], dfpackets['diry'][pkt_index], dfpackets['dirz'][pkt_index]]
        costheta = at.dot(pkt_dir, syn_dir)
        thetabin = ((costheta + 1.0) * np.sqrt(MABINS) / 2.0)
        vec1 = vec2 = vec3 = [0, 0, 0]
        xhat = [1, 0, 0]
        vec1 = at.cross_prod(pkt_dir, syn_dir, vec1)
        vec2 = at.cross_prod(xhat, syn_dir, vec2)
        cosphi = at.dot(vec1, vec2) / at.vec_len(vec1) / at.vec_len(vec2)

        vec3 = at.cross_prod(vec2, syn_dir, vec3)
        testphi = at.dot(vec1, vec3)

        if testphi > 0:
            phibin = (math.acos(cosphi) / 2. / np.pi * np.sqrt(MABINS))
        else:
            phibin = ((math.acos(cosphi) + np.pi) / 2. / np.pi * np.sqrt(MABINS))
        na = (thetabin * np.sqrt(MABINS)) + phibin  # think na is angle number???
        if na >= 100:
            print(f'error bin number too high {na}')
        angle_number[i] = int(na)
        i += 1

    dfpackets['angle_bin'] = angle_number
    return dfpackets


def make_3d_histogram_from_packets(modelpath, timestep):
    modeldata, _, vmax_cms = at.inputmodel.get_modeldata(modelpath)

    timeminarray = at.get_timestep_times_float(modelpath=modelpath, loc='start')
    timedeltaarray = at.get_timestep_times_float(modelpath=modelpath, loc='delta')
    timemaxarray = at.get_timestep_times_float(modelpath=modelpath, loc='end')

    # timestep = 63 # 82 73 #63 #54 46 #27
    print([(ts, time) for ts, time in enumerate(timeminarray)])

    packetsfiles = at.packets.get_packetsfilepaths(modelpath)

    emission_position3d = [[], [], []]
    e_rf = []

    for npacketfile in range(0, len(packetsfiles)):
        # for npacketfile in range(0, 1):
        dfpackets = at.packets.readfile(packetsfiles[npacketfile])
        at.packets.add_derived_columns(dfpackets, modelpath, ['emission_velocity'])
        dfpackets = dfpackets.dropna(subset=['emission_velocity'])  # drop rows where emission_vel is NaN

        # print(dfpackets[['emission_velocity', 'em_velx', 'em_vely', 'em_velz']])
        # select only type escape and type r-pkt (don't include gamma-rays)
        dfpackets.query(f'type_id == {type_ids["TYPE_ESCAPE"]} and escape_type_id == {type_ids["TYPE_RPKT"]}',
                        inplace=True)
        dfpackets.query('@timeminarray[@timestep] < escape_time/@DAY < @timemaxarray[@timestep]', inplace=True)

        emission_position3d[0].extend([em_velx / CLIGHT for em_velx in dfpackets['em_velx']])
        emission_position3d[1].extend([em_vely / CLIGHT for em_vely in dfpackets['em_vely']])
        emission_position3d[2].extend([em_velz / CLIGHT for em_velz in dfpackets['em_velz']])

        e_rf.extend([e_rf for e_rf in dfpackets['e_rf']])

    emission_position3d = np.array(emission_position3d)
    weight_by_energy = False
    if weight_by_energy:
        e_rf = np.array(e_rf)
        weights = e_rf
    else:
        weights = None

    print(emission_position3d.shape)
    print(emission_position3d[0].shape)

    # print(emission_position3d)
    grid_3d, _, _, _ = make_3d_grid(modeldata, vmax_cms)
    print(grid_3d)
    # https://stackoverflow.com/questions/49861468/binning-random-data-to-regular-3d-grid-with-unequal-axis-lengths
    hist, _ = np.histogramdd(emission_position3d.T, [np.append(ax, np.inf) for ax in grid_3d], weights=weights)
    # print(hist.shape)
    if weight_by_energy:
        # Divide binned energies by number of processes and by length of timestep
        hist = hist / len(packetsfiles) / timedeltaarray[timestep]  # histogram weighted by energy
    # - need to divide by number of processes
    # and length of timestep

    # # print histogram coordinates
    # coords = np.nonzero(hist)
    # for i, j, k in zip(*coords):
    #     print(f'({grid_3d[0][i]}, {grid_3d[1][j]}, {grid_3d[2][k]}): {hist[i][j][k]}')

    return hist


def make_3d_grid(modeldata, vmax_cms):
    # modeldata, _, vmax_cms = at.inputmodel.get_modeldata(modelpath)
    grid = round(len(modeldata['inputcellid']) ** (1. / 3.))
    xgrid = np.zeros(grid)
    vmax = vmax_cms / CLIGHT
    i = 0
    for z in range(0, grid):
        for y in range(0, grid):
            for x in range(0, grid):
                xgrid[x] = -vmax + 2 * x * vmax / grid
                i += 1

    x, y, z = np.meshgrid(xgrid, xgrid, xgrid)
    grid_3d = np.array([xgrid, xgrid, xgrid])
    # grid_Te = np.zeros((grid, grid, grid))
    # print(grid_Te.shape)
    return grid_3d, x, y, z


def get_mean_packet_emission_velocity_per_ts(modelpath, packet_type='TYPE_ESCAPE', escape_type='TYPE_RPKT',
                                             maxpacketfiles=None, escape_angles=None):

    packetsfiles = at.packets.get_packetsfilepaths(modelpath, maxpacketfiles=maxpacketfiles)
    nprocs_read = len(packetsfiles)
    assert nprocs_read > 0

    timearray = at.get_timestep_times_float(modelpath=modelpath, loc='mid')
    arr_timedelta = at.get_timestep_times_float(modelpath=modelpath, loc='delta')
    timearrayplusend = np.concatenate([timearray, [timearray[-1] + arr_timedelta[-1]]])

    dfpackets_escape_velocity_and_arrive_time = pd.DataFrame
    emission_data = pd.DataFrame({'t_arrive_d': timearray,
                                  'mean_emission_velocity': np.zeros_like(timearray, dtype=float)})

    for i, packetsfile in enumerate(packetsfiles):
        dfpackets = at.packets.readfile(packetsfile, type=packet_type, escape_type=escape_type)
        at.packets.add_derived_columns(dfpackets, modelpath, ['emission_velocity'])
        if escape_angles is not None:
            dfpackets = at.packets.get_escaping_packet_angle_bin(modelpath, dfpackets)
            dfpackets.query('angle_bin == @escape_angles', inplace=True)

        if i == 0:  # make new df
            dfpackets_escape_velocity_and_arrive_time = dfpackets[['t_arrive_d', 'emission_velocity']]
        else:  # append to df
            dfpackets_escape_velocity_and_arrive_time = dfpackets_escape_velocity_and_arrive_time.append(
                dfpackets[['t_arrive_d', 'emission_velocity']], ignore_index=True)

    print(dfpackets_escape_velocity_and_arrive_time)
    binned = pd.cut(dfpackets_escape_velocity_and_arrive_time['t_arrive_d'], timearrayplusend,
                    labels=False, include_lowest=True)
    for binindex, emission_velocity in dfpackets_escape_velocity_and_arrive_time.groupby(binned)[
            'emission_velocity'].mean().iteritems():
        emission_data['mean_emission_velocity'][binindex] += emission_velocity  # / 2.99792458e10

    return emission_data



#!/usr/bin/env python3

import argparse
# import math
# import os.path

from astropy import units as u
# import numpy as np
# import pandas as pd
from pathlib import Path

import artistools as at


def addargs(parser):
    parser.add_argument('-scalefactor', '-s',
                        default=0.5,
                        help='Kinetic energy scale factor')
    parser.add_argument('-inputpath', '-i',
                        default='.',
                        help='Path of input files')
    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path of output files')


def eval_mshell(dfmodel, t_model_init_seconds):
    dfmodel.eval('cellmass_grams = 10 ** logrho * 4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                     '* (1e5 * @t_model_init_seconds) ** 3', inplace=True)


def main(args=None, argsraw=None, **kwargs) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Fully mix an ARTIS model to homogenous composition and save back to ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    dfmodel, t_model_init_days, _ = at.inputmodel.get_modeldata(args.inputpath)
    print('Read model.txt')
    dfelabundances = at.inputmodel.get_initialabundances(args.inputpath)
    print('Read abundances.txt')

    t_model_init_seconds = t_model_init_days * 24 * 60 * 60

    eval_mshell(dfmodel, t_model_init_seconds)

    print(dfmodel)
    print(dfelabundances)

    model_mass_grams = dfmodel.cellmass_grams.sum()
    print(f"model mass: {model_mass_grams * u.g.to('solMass'):.3f} Msun")
    for column_name in [x for x in dfmodel.columns if x.startswith('X_')]:
        integrated_mass_grams = (dfmodel[column_name] * dfmodel.cellmass_grams).sum()
        global_massfrac = integrated_mass_grams / model_mass_grams
        print(f"{column_name:>13s}: {global_massfrac:.3f}  ({integrated_mass_grams * u.g.to('solMass'):.3f} Msun)")
        dfmodel.eval(f'{column_name} = {global_massfrac}', inplace=True)

    for column_name in [x for x in dfelabundances.columns if x.startswith('X_')]:
        integrated_mass_grams = (dfelabundances[column_name] * dfmodel.cellmass_grams).sum()
        global_massfrac = integrated_mass_grams / model_mass_grams
        print(f"{column_name:>13s}: {global_massfrac:.3f}  ({integrated_mass_grams * u.g.to('solMass'):.3f} Msun)")
        dfelabundances.eval(f'{column_name} = {global_massfrac}', inplace=True)

    print(dfmodel)
    print(dfelabundances)

    modeloutfilename = "model_fullymixed.txt"
    at.save_modeldata(dfmodel, t_model_init_days, Path(args.outputpath, modeloutfilename))
    print(f'Saved {modeloutfilename}')

    abundoutfilename = "abundances_fullymixed.txt"
    at.inputmodel.save_initialabundances(dfelabundances, Path(args.outputpath, abundoutfilename))
    print(f'Saved {abundoutfilename}')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import multiprocessing
import os
import pynonthermal
from functools import lru_cache
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
# import matplotlib.ticker as ticker
import pandas as pd
from astropy import units as u

import artistools as at
# from artistools.nonthermal.spencerfano import *

DEFAULTSPECPATH = '../example_run/spec.out'
defaultoutputfile = 'plotnonthermal_cell{0:03d}_timestep{1:03d}.pdf'


@lru_cache(maxsize=4)
def read_files(modelpath, timestep=-1, modelgridindex=-1):
    """Read ARTIS -thermal spectrum data into a pandas DataFrame."""
    nonthermaldata = pd.DataFrame()

    mpiranklist = at.get_mpiranklist(modelpath, modelgridindex=modelgridindex)
    for folderpath in at.get_runfolders(modelpath, timestep=timestep):
        for mpirank in mpiranklist:
            nonthermalfile = f'nonthermalspec_{mpirank:04d}.out'
            filepath = at.firstexisting([nonthermalfile, nonthermalfile + '.xz', nonthermalfile + '.gz'], path=folderpath)

            if modelgridindex > -1:
                filesize = Path(filepath).stat().st_size / 1024 / 1024
                print(f'Reading {Path(filepath).relative_to(modelpath.parent)} ({filesize:.2f} MiB)')

            nonthermaldata_thisfile = pd.read_csv(filepath, delim_whitespace=True, on_bad_lines='skip')
            # radfielddata_thisfile[['modelgridindex', 'timestep']].apply(pd.to_numeric)

            if timestep >= 0:
                nonthermaldata_thisfile.query('timestep==@timestep', inplace=True)

            if modelgridindex >= 0:
                nonthermaldata_thisfile.query('modelgridindex==@modelgridindex', inplace=True)

            if not nonthermaldata_thisfile.empty:
                if timestep >= 0 and modelgridindex >= 0:
                    return nonthermaldata_thisfile
                else:
                    nonthermaldata = nonthermaldata.append(nonthermaldata_thisfile.copy(), ignore_index=True)

    return nonthermaldata


def make_xs_plot(axis, nonthermaldata, args):
    dfcollion = at.nonthermal.read_colliondata()

    arr_en = nonthermaldata['energy_ev'].unique()

    # arr_xs_old = [xs_fe2_old(en) for en in arr_en]
    # arr_xs_times_y = [xs_fe1(en) * y for en, y in zip(nonthermaldata['energy_ev'], nonthermaldata['y'])]

    axis.plot(arr_en, pynonthermal.collion.get_arxs_array_ion(arr_en, dfcollion, 26, 2), linewidth=2.0, label='Fe II')
    axis.plot(arr_en, pynonthermal.collion.get_arxs_array_ion(arr_en, dfcollion, 28, 2), linewidth=2.0, label='Ni II')

    axis.set_ylabel(r'cross section (cm2)')

    if not args.nolegend:
        axis.legend(loc='upper center', handlelength=2, frameon=False, numpoints=1, prop={'size': 13})


def inteuler(x, y):
    dx = y[1:] - y[:-1]
    return np.dot(x[:-1], dx)


def plot_contributions(axis, modelpath, timestep, modelgridindex, nonthermaldata, args):
    estimators = at.estimators.read_estimators(modelpath, get_ion_values=True, get_heatingcooling=True,
                                               modelgridindex=modelgridindex, timestep=timestep)

    # print(estimators[(timestep, modelgridindex)].keys())
    total_depev = (estimators[(timestep, modelgridindex)]['total_dep'] * u.erg.to('eV'))

    print(f"Deposition: {total_depev:.1f} [eV/cm3/s]")

    arr_enev = nonthermaldata['energy_ev'].values
    arr_y = nonthermaldata['y'].values

    frac_ionisation = 0.

    dfcollion = at.nonthermal.read_colliondata()

    elementlist = at.get_composition_data(modelpath)
    totalpop = estimators[(timestep, modelgridindex)]['populations']['total']
    nelements = len(elementlist)
    for element in range(nelements):
        Z = elementlist.Z[element]

        elpop = estimators[(timestep, modelgridindex)]['populations'][Z]
        if elpop <= 1e-4 * totalpop:
            continue

        arr_ionisation_element = np.zeros(len(arr_enev), dtype=float)
        frac_ionisation_element = 0.

        nions = elementlist.nions[element]
        for ion in range(nions):
            ionstage = ion + elementlist.lowermost_ionstage[element]
            ionpop = estimators[(timestep, modelgridindex)]['populations'][(Z, ionstage)]

            dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage')

            # print(at.get_ionstring(Z, ionstage), ionpop)

            arr_ionisation_ion = np.zeros(len(arr_enev), dtype=float)
            frac_ionisation_ion = 0.

            for index, row in dfcollion_thision.iterrows():
                arr_xs = at.nonthermal.get_arxs_array_shell(arr_enev, row)
                arr_ionisation_shell = ionpop * arr_y * arr_xs * row.ionpot_ev / total_depev
                arr_ionisation_ion += arr_ionisation_shell

                frac_ionisation_shell = np.trapz(x=arr_enev, y=arr_ionisation_shell)
                frac_ionisation_ion += frac_ionisation_shell

            arr_ionisation_element += arr_ionisation_ion
            frac_ionisation_element += frac_ionisation_ion

        frac_ionisation += frac_ionisation_element

        if frac_ionisation_element > 1e-5:
            axis.plot(arr_enev, arr_ionisation_element, label=f'Ionisation Z={Z}')

    nne = estimators[(timestep, modelgridindex)]['nne']
    arr_heating = np.array([at.nonthermal.lossfunction(enev, nne) / total_depev for enev in arr_enev])

    frac_heating = np.trapz(x=arr_enev, y=arr_heating)

    print(f'   frac_heating: {frac_heating}')
    print(f'frac_ionisation: {frac_ionisation}')

    axis.plot(arr_enev, arr_heating, label='Heating')

    axis.legend(loc='best', handlelength=2, frameon=False, numpoints=1, prop={'size': 11})


def make_plot(modelpaths, args):
    nplots = 1
    if args.xsplot:
        nplots += 1
    if args.showcontributions:
        nplots += 1
    fig, axes = plt.subplots(nrows=nplots, ncols=1, sharex=True,
                             figsize=(args.figscale * at.figwidth, args.figscale * at.figwidth * 0.7 * nplots),
                             tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    if nplots == 1:
        axes = [axes]

    if args.kf1992spec:
        kf92spec = pd.read_csv(Path(modelpaths[0], 'KF1992spec-fig1.txt'), header=None, names=['e_kev', 'log10_y'])
        kf92spec['energy_ev'] = kf92spec['e_kev'] * 1000.
        kf92spec.eval('y = 10 ** log10_y', inplace=True)
        axes[0].plot(kf92spec['energy_ev'], kf92spec['log10_y'],
                     linewidth=2.0, color='red', label='Kozma & Fransson (1992)')

    for index, modelpath in enumerate(modelpaths):
        modelname = at.get_model_name(modelpath)
        if args.velocity >= 0.:
            modelgridindex = at.get_mgi_of_velocity_kms(modelpath, args.velocity)
        else:
            modelgridindex = args.modelgridindex

        if args.timedays:
            timestep = at.get_timestep_of_timedays(modelpath, args.timedays)
        else:
            timestep = args.timestep

        nonthermaldata = read_files(
            modelpath=Path(modelpath),
            modelgridindex=modelgridindex, timestep=timestep)

        if args.xmin:
            nonthermaldata.query('energy_ev >= @args.xmin', inplace=True)

        if nonthermaldata.empty:
            print(f'No data for timestep {timestep:d}')
            continue

        if index < len(args.modellabels):
            model_label = args.modellabels[index]
        else:
            model_label = f'{modelname} cell {modelgridindex} at timestep {timestep}'
            try:
                time_days = float(at.get_timestep_time('.', timestep))
            except FileNotFoundError:
                time_days = 0
            else:
                model_label += f' ({time_days:.2f}d)'

        outputfile = str(args.outputfile).format(modelgridindex, timestep)
        print(f'Plotting timestep {timestep:d}')
        # ymax = max(nonthermaldata['y'])

        # nonthermaldata.plot(x='energy_ev', y='y', linewidth=1.5, ax=axis, color='blue', legend=False)
        axes[0].plot((nonthermaldata['energy_ev']), np.log10(nonthermaldata['y']), label=model_label,
                     linewidth=2.0, color='black' if index == 0 else None, alpha=0.95)
        axes[0].set_ylabel(r'log [y (e$^-$ / cm$^2$ / s / eV)]')

        if args.showcontributions:
            plot_contributions(axes[1], modelpath, timestep, modelgridindex, nonthermaldata, args)

        if args.xsplot:
            make_xs_plot(axes[-1], nonthermaldata, args)

    if not args.nolegend:
        axes[0].legend(loc='best', handlelength=2, frameon=False, numpoints=1)

    axes[-1].set_xlabel(r'Energy (eV)')
    # axis.yaxis.set_minor_locator(ticker.MultipleLocator(base=0.1))
    # axis.set_yscale("log", nonposy='clip')
    for ax in axes:
        if args.xmin is not None:
            ax.set_xlim(left=args.xmin)
        if args.xmax:
            ax.set_xlim(right=args.xmax)
    # axis.set_ylim(bottom=0.0, top=ymax)

    # axis.legend(loc='upper center', handlelength=2,
    #             frameon=False, numpoints=1, prop={'size': 13})

    print(f'Saving to {outputfile:s}')
    fig.savefig(outputfile, format='pdf')
    plt.close()


def addargs(parser):
    parser.add_argument('-modelpath', default=[], nargs='*', action=at.AppendPath,
                        help='Paths to ARTIS folders with spec.out or packets files')

    parser.add_argument('-modellabels', default=[], nargs='*',
                        help='Model name overrides')

    parser.add_argument('-listtimesteps', action='store_true',
                        help='Show the times at each timestep')

    parser.add_argument('-xsplot', action='store_true',
                        help='Show the cross section plot')

    parser.add_argument('-timedays', '-time', '-t',
                        help='Time in days to plot')

    parser.add_argument('-timestep', '-ts', type=int, default=-1,
                        help='Timestep number to plot')

    parser.add_argument('-modelgridindex', '-cell', type=int, default=0,
                        help='Modelgridindex to plot')

    parser.add_argument('-velocity', '-v', type=float, default=-1,
                        help='Specify cell by velocity')

    parser.add_argument('-xmin', type=float, default=0.,
                        help='Plot range: minimum energy in eV')

    parser.add_argument('-xmax', type=float,
                        help='Plot range: maximum energy in eV')

    parser.add_argument('--nolegend', action='store_true',
                        help='Suppress the legend from the plot')

    parser.add_argument('--showcontributions', action='store_true',
                        help='Plot the NT contributions to ionisation and heating energy')

    parser.add_argument('--kf1992spec', action='store_true',
                        help='Show the pure-oxygen result form Figure 1 of Kozma & Fransson 1992')

    parser.add_argument('-figscale', type=float, default=1.,
                        help='Scale factor for plot area. 1.0 is for single-column')

    parser.add_argument('-o', action='store', dest='outputfile', type=Path,
                        default=defaultoutputfile,
                        help='Filename for PDF file')


def main(args=None, argsraw=None, **kwargs):
    """Plot the electron energy distribution."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot ARTIS non-thermal electron energy spectrum.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = [Path('.')]
    elif isinstance(args.modelpath, (str, Path)):
        args.modelpath = [args.modelpath]

    # flatten the list
    modelpaths = []
    for elem in args.modelpath:
        if isinstance(elem, list):
            modelpaths.extend(elem)
        else:
            modelpaths.append(elem)

    if os.path.isdir(args.outputfile):
        args.outputfile = os.path.join(args.outputfile, defaultoutputfile)

    if args.listtimesteps:
        at.showtimesteptimes()
    else:
        make_plot(modelpaths, args)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



import argcomplete
import argparse
import importlib

import artistools.commands


def addargs(parser):
    pass


def main(argsraw=None):
    """Show a list of available artistools commands."""

    parser = argparse.ArgumentParser()
    parser.set_defaults(func=None)

    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = False

    for command, (submodulename, funcname) in sorted(artistools.commands.commandlist.items()):
        submodule = importlib.import_module(submodulename, package='artistools')
        subparser = subparsers.add_parser(command.replace('artistools-', ''))
        submodule.addargs(subparser)
        subparser.set_defaults(func=getattr(submodule, funcname))

    argcomplete.autocomplete(parser)
    args = parser.parse_args(argsraw)
    if args.func is not None:
        args.func(args=args)
    else:
        # parser.print_help()
        print('artistools provides the following commands:\n')

        # for script in sorted(console_scripts):
        #     command = script.split('=')[0].strip()
        #     print(f'  {command}')

        for command in sorted(artistools.commands.commandlist):
            print(f'  {command}')


if __name__ == '__main__':
    # multiprocessing.freeze_support()
    main()



#!/usr/bin/env python3

import argparse
import math
# import os.path

# import numpy as np
# import pandas as pd

import artistools as at


def addargs(parser):

    parser.add_argument('-kescale', '-k', default=None,
                        help='Kinetic energy scale factor')

    parser.add_argument('-velscale', '-v', default=None,
                        help='Velocity scale factor')

    parser.add_argument('-inputfile', '-i', default='model.txt',
                        help='Path of input file')

    parser.add_argument('-outputfile', '-o', default='model_velscale{velscale:.2f}.txt',
                        help='Path of output model file')


def eval_mshell(dfmodel, t_model_init_seconds):
    dfmodel.eval('cellmass_grams = 10 ** logrho * 4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                     '* (1e5 * @t_model_init_seconds) ** 3', inplace=True)


def main(args=None, argsraw=None, **kwargs) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Scale the velocity of an ARTIS model, keeping mass constant and saving back to ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    dfmodel, t_model_init_days, _ = at.inputmodel.get_modeldata(args.inputfile)
    print(f'Read {args.inputfile}')

    t_model_init_seconds = t_model_init_days * 24 * 60 * 60

    eval_mshell(dfmodel, t_model_init_seconds)

    print(dfmodel)

    assert (args.kescale is None) != (args.velscale is None)  # kescale or velscale must be specfied

    if args.kescale is not None:
        kescale = float(args.kescale)
        velscale = math.sqrt(kescale)
    elif args.velscale is not None:
        velscale = float(args.velscale)
        kescale = velscale ** 2

    print(f"Applying velocity factor of {velscale} (kinetic energy factor {kescale}) and conserving shell masses")

    dfmodel.velocity_inner *= velscale
    dfmodel.velocity_outer *= velscale

    dfmodel.eval('logrho = log10(cellmass_grams / ('
                     '4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                     ' * (1e5 * @t_model_init_seconds) ** 3))', inplace=True)

    eval_mshell(dfmodel, t_model_init_seconds)

    print(dfmodel)

    outputfile = args.outputfile.format(velscale=velscale, kescale=kescale)

    at.inputmodel.save_modeldata(dfmodel=dfmodel, t_model_init_days=t_model_init_days, filename=outputfile)
    print(f'Saved {outputfile}')


if __name__ == "__main__":
    main()
